;;;*************************************************************************
;;;module name: semprop.sch
;;;author     : Reinhold Letz, TU Munich
;;;date       : 29.10.98
;;;description: propositional decision procedure using the semantic tree 
;;;		calculus (DPLL-kernel)
;;;remarks    : literals are integers and must not occur multiply in a clause.
;;;compiling  : yes
;;;*************************************************************************

;; -------------------------------------------------------------------------
;; internal data structures:
;; 
;; clause: vector of four elements
;; 1st element: atom which activates the clause [#f or atom]
;; 2nd element: current clause length [positive integer]
;; 3rd element: list of positive atoms
;; 4th element: list of negative atoms
;; 5th element: clause number
;;
;; atom: vector of seven elements
;; 1st element: atom is currently unevaluated [#f] or evaluated [0 or 1] 
;; 2nd element: clauses with positive occurrence of atom
;; 3rd element: clauses with negative occurrence of atom
;; 4th element: last inference using assignment of atom
;; 5th element: positive atom frequency in formula
;; 6th element: negative atom frequency in formula
;; 7th element: atom name
;; 8th element: atom sign
;; 
;; -------------------------------------------------------------------------

;; macros for accessing and modifying basic data structures
;;
(define-macro (c-activating-atom clause) `(vector-ref ,clause 0))
(define-macro (c-length clause)          `(vector-ref ,clause 1))
(define-macro (c-pos-literals clause)    `(vector-ref ,clause 2))
(define-macro (c-neg-literals clause)    `(vector-ref ,clause 3))
(define-macro (c-number clause)          `(vector-ref ,clause 4))

(define-macro (set-c-activating-atom clause val) `(vector-set! ,clause 0 ,val))
(define-macro (set-c-length clause val)          `(vector-set! ,clause 1 ,val))
(define-macro (set-c-pos-literals clause val)    `(vector-set! ,clause 2 ,val))
(define-macro (set-c-neg-literals clause val)    `(vector-set! ,clause 3 ,val))
(define-macro (set-c-number clause val)          `(vector-set! ,clause 4 ,val))

(define-macro (a-evaluated atom)         `(vector-ref ,atom 0))
(define-macro (a-pos-clauses atom)       `(vector-ref ,atom 1))
(define-macro (a-neg-clauses atom)       `(vector-ref ,atom 2))
(define-macro (a-last-inference	atom)    `(vector-ref ,atom 3))
(define-macro (a-pos-frequency atom)     `(vector-ref ,atom 4))
(define-macro (a-neg-frequency atom)     `(vector-ref ,atom 5))
(define-macro (a-name atom)              `(vector-ref ,atom 6))
(define-macro (a-sign atom)              `(vector-ref ,atom 7))

(define-macro (set-a-evaluated atom val)      `(vector-set! ,atom 0 ,val))
(define-macro (set-a-pos-clauses atom val)    `(vector-set! ,atom 1 ,val))
(define-macro (set-a-neg-clauses atom val)    `(vector-set! ,atom 2 ,val))
(define-macro (set-a-last-inference atom val) `(vector-set! ,atom 3 ,val))
(define-macro (set-a-pos-frequency atom val)  `(vector-set! ,atom 4 ,val))
(define-macro (set-a-neg-frequency atom val)  `(vector-set! ,atom 5 ,val))
(define-macro (set-a-name atom val)           `(vector-set! ,atom 6 ,val))
(define-macro (set-a-sign atom val)           `(vector-set! ,atom 7 ,val))

(define-macro (branching atom-index atom sign inference)
;; a proper semantic tree branching is performed
`(let (clauses true-clauses new-units new-inference1 new-inference2)
 (newline) (display "INFERENCE ") (display inference) (display " BRANCHING: ") ;d;
  (set! ,atom (select-atom ,atom-index))
  (set-a-evaluated ,atom 0) ;; atom is made unavailable in left mode
  (if (< (a-pos-frequency ,atom) (a-neg-frequency ,atom))
      (begin (set! ,sign 0) (set! clauses (a-pos-clauses ,atom))
             (set-a-sign ,atom 0)
 (display "-") ;d;
             (set! true-clauses (a-neg-clauses ,atom)))
      (begin (set! ,sign 1) (set! clauses (a-neg-clauses ,atom))
             (set-a-sign ,atom 1)
             (set! true-clauses (a-pos-clauses ,atom))))
 (display (a-name ,atom)) ;d;
(newline) (display unit-index) ;d;
  (decr-counters clauses)
  (disactivate true-clauses ,atom)
 (newline) (display-all-clauses clause-list) ;d;
  (find-active) (set! new-inference1 (evaluate ,atom-index ,inference))
  (reactivate true-clauses ,atom) ;; clauses are reactivated
  (incr-counters clauses) ;; clause lengths are incremented
  (if new-inference1 ; no model was found in the left subtree
      (if (< (a-last-inference ,atom) ,inference) ; literal not used in subtree
        (begin (set-a-evaluated ,atom #f);; atom is made available
               (set-a-sign ,atom #f) 
               new-inference1)
        (begin (set! unit-index -1)
          (decr-counters true-clauses)
          (set-a-evaluated ,atom 1) ;; atom is made unavailable in right mode
          (disactivate clauses ,atom)
          (if (= ,sign 1) (set! ,sign 0) (set! ,sign 1))
          (find-active)
          (set! new-inference2 (evaluate ,atom-index new-inference1))
          (set-a-evaluated ,atom #f);; atom is made available
          (set-a-sign ,atom #f) 
          (reactivate clauses ,atom) ;; clauses are reactivated
          (incr-counters true-clauses) ;; clause lengths are incremented
          new-inference2))
      (begin (set-a-evaluated ,atom #f);; atom is made available
             (set-a-sign ,atom #f) 
             #f))))

(define-macro (make-atom atom-name) `(vector #f '() '() 0 0 0 ,atom-name #f))

(define-macro (select-atom atom-index) `(vector-ref atom-vector ,atom-index))

