/******************************************************/
/*    S E T H E O                                     */
/*                                                    */
/* FILE: i_equconstr.c                                */
/* VERSION:                                           */
/* DATE:                                              */
/* AUTHOR:                                            */
/* NAME OF FILE:                                      */
/* DESCR:                                             */
/* MOD:                                               */
/* BUGS:                                              */
/******************************************************/

#include <stdio.h>
#include "tags.h"
#include "machine.h"
#include "deref.h"
#include "errors.h"
#include "disp.h"
#include "unitlemma.h"
#include "sem_constr.h"
#include "sem_tools.h"
#include "symbtab.h"
#include "univ_constr.h"

#define INSTR_LENGTH            2
#define LOCAL_DEBUG2
#define LOCAL_DEBUG2


/* $equconstr/2 creates a constraint tree

   - A junctor is one of the predicates "and" and "or" with variable arity.
     Each argument is either a junctor or an elementary constraint.
   - An elementary constraint is a list of two lists (of the same length

   An example of a junctor is
   sem_or( sem_and( [[X, Y], [1, 0]] ), [[Z], [1]])

   Any combination is possible and only a few checks are being done.

   The second argument of this builtin is the number of clauses that
   (during preprocessing) could be summarized with this constraint.
   As there`s only one root junctor, $equconstr is binary.
*/



/* self is a pointer to a list that encodes elementary constraint.
   Parse it and attach the corresponding elem constrs to
   parentJunct.
   */
int  parse_list_equation(parentJunct, self)
sem_junct*  parentJunct;
WORD*       self;
{
   WORD*       leftList;
   WORD*       rightList;
   int         listSuccess = FALSE;
   
   /* self has to be a list with two elements, that are nonempty lists themselves
      Extract these two lists and write them into leftList and rightList */
   extract_constraint_sides(self, &leftList, &rightList);
   
   if (list_length(leftList) != list_length(rightList))
      sam_error("equconstr: The constraint-sides are unbalanced",
		NULL, 2);
   
#ifdef LOCAL_DEBUG1
   printf("processing list equation ...\n");
   printf("leftList: ");
   disp_(stdout, leftList, bp);
   printf("\nrightList: ");
   disp_(stdout, rightList, bp);
   printf("\n");
   fflush(stdout);
#endif

   pu_tr = pu_trail;
   
   if (un_protoco(leftList, bp, rightList, bp, 1)) {
      if (pu_tr > pu_trail) {
	 generate_sem_elem_constraints_from_pu_trail(parentJunct,
						     ELEM_EQU,
						     FALSE);
      }
      listSuccess = TRUE;
   }
   else
      pu_untrail();

#ifdef LOCAL_DEBUG1
   printf("\nparent of list equation:");
   print_sem_constr_tree(parentJunct, 0);
   printf("\n\n");
   fflush(stdout);
#endif

   return listSuccess;
}



/* self points to a predicate that encodes a junctor
   Parse this predicate, add the corresponding sem_junct to parentJunct
   and return it afterwards
   */
sem_junct*  parse_junct(parentJunct, self)
sem_junct*  parentJunct;
WORD*       self;
{
   sem_junct*  outJunct;
   WORD*       curArg;
   int         count;
   int         junctSuccess = 0;
   
#ifdef LOCAL_DEBUG1
   printf("processing junctor ...\n");
#endif
   
  /* Check junction symbol: */
   switch (GETSYMBOLINDEX(*self)) {
   case AND_JUNCT:
	 if (parentJunct
             &&  parentJunct->type == JUNCT_CONJ)
	    outJunct = parentJunct;
	 else
	    outJunct = create_sem_junct(JUNCT_CONJ);
	 break;
   case OR_JUNCT:
	 if (parentJunct
             &&  parentJunct->type == JUNCT_DISJ)
	    outJunct = parentJunct;
	 else
	    outJunct = create_sem_junct(JUNCT_DISJ);
	 break;
   default:
	 sam_error("equconstr: Wrong predicate in constraint tree.", 
		   NULL, 2);
   }
   
   /* Add the arguments of the predicate to the junctor: */
   for ( count = 1; count <= GETARITY(*self); count++ ) {

      curArg = deref(self+count, bp);

      switch (GETSYMBOLINDEX(*curArg)) {
      case LISTFUNCT:
	    if (parse_list_equation(outJunct, curArg))
	       junctSuccess++;
	    break;
      case AND_JUNCT:
      case OR_JUNCT:
	    if (parse_junct(outJunct, curArg))
	       junctSuccess++;
	    break;
      default:
	    sam_error("equconstr: Wrong construct in constraint tree",
		      NULL, 2);
      }


      /* Remove AND junctor and return NULL, if the conjunction is not
	 satisfied: */
      if (GETSYMBOLINDEX(*self) == AND_JUNCT  &&  junctSuccess != count) {
	 if (parentJunct  &&  outJunct != parentJunct)
	    c_sp = (WORD*) outJunct;
	 return NULL;
      }
   }

   /* Remove OR junctor and return NULL, if the disjunction is not
      satisfied: */
   if (GETSYMBOLINDEX(*self) == OR_JUNCT  &&  junctSuccess < 1) {
      if (parentJunct  &&  outJunct != parentJunct)
	 c_sp = (WORD*) outJunct;
      return NULL;
   }
   
   /* If outJunct is not the root junctor 
      and if outJunct is not identical to its parent junctor: 
      - if outJunct is trivially satisfied (arity==zero), 
        remove it and return something != NULL
      - otherwise
        - add outJunct to its parent
	- if outJunct has arity==1, 
	  remove outJunct from the constraint tree */
   if (parentJunct  &&  outJunct != parentJunct) {
      if (outJunct->arity == 0) {
	 c_sp = (WORD*) outJunct;
	 return outJunct;
      }
      else {
	 add_sem_junct(parentJunct, outJunct);
	 if (outJunct->arity == 1)
	    remove_junct_with_arity_one(outJunct);
      }
   }

#ifdef LOCAL_DEBUG1
   printf("\ngenerated junctor (end of parse junct):");
   print_sem_constr_tree(outJunct, 0);
   printf("\n\n");
   fflush(stdout);
#endif
   
   /* The return value is only used if outJunct is the root junctor: */
   return outJunct;
}



instr_result i_equconstr()
{
   WORD*       topArg;
   sem_junct*  rootJunct;
   
   topArg = deref(ARGPTR(1), bp);

#ifdef LOCAL_DEBUG
   printf("Processing constraints of clause %d ... ", 
	  GETNUMBER(bp->lcbptr->clause_nr));
#endif
   
   /* The rootJunct is a sem_junct whose parent is NULL
      (This call creates the constraint-tree.) */
   rootJunct = parse_junct(NULL, topArg);

   if (!rootJunct) {
      chp->consistency_fail_occurred = TRUE;
#ifdef LOCAL_DEBUG
      printf("not successful\n");
      fflush(stdout);
#endif
      return failure;
   }
   
#ifdef LOCAL_DEBUG1
   printf("generated rootJunct:");
   print_sem_constr_tree(rootJunct, 0);
   printf("\n\n");
#endif

   if (rootJunct->arity == 0) {
      c_sp = (WORD*) rootJunct;
   }
   else {
      insert_in_elem_constr_lists(rootJunct);
      attach_to_imagined_root_node(rootJunct);

      /* try deterministic instantiation: */
      if (   (     rootJunct->type == JUNCT_CONJ
	      ||  (rootJunct->type == JUNCT_DISJ  &&  rootJunct->arity == 1))
	  &&  try_det_inst() == 0) {
	 /* chp->consistency_fail_occurred = TRUE; 
	    need not be set here because is always set when
	    deterministic instantiation is called */
#ifdef LOCAL_DEBUG
         printf("not successful\n");
         fflush(stdout);
#endif
	 return failure;
      }

      /* If the root node is still alive, indicate that a new
	 constraint tree has been added (useful for selective local
	 consistency check in call instruction): */
      if (is_sem_junct_alive(rootJunct)) {
	 sem_constr_added = TRUE;  /* flag, used for control of local
				      consistency check */
	 constr_tree_count++;      /* counter, used for control of local
				      consistency check */
      }
   }
   
#ifdef LOCAL_DEBUG
   printf("successful\n");
   fflush(stdout);
#endif
#ifdef LOCAL_DEBUG1
   printf("\n\nnewsemantic_root:");
   print_sem_constr_tree(semantic_root, 0);
   printf("\n\n");
   fflush(stdout);
#endif
   pc +=INSTR_LENGTH;
   return success;
}
