/******************************************************/
/*    S E T H E O                                     */
/*                                                    */
/* FILE: sem_consistency.c                            */
/* VERSION:                                           */
/* DATE:                                              */
/* AUTHOR: Rauschmayer & Sacher                       */
/* NAME OF FILE:                                      */
/* DESCR:                                             */
/* MOD:                                               */
/* BUGS:                                              */
/******************************************************/

#include <stdio.h>
#include <math.h>
#include "tags.h"
#include "machine.h"
#include "symbtab.h"
#include "deref.h"
#include "errors.h"
#include "proto_unify.h"
#include "sem_consistency.h"
#include "sem_constr.h"
#include "sem_tools.h"
#include "unification.h"
#include "unifeq.h"
#include "unitlemma.h"

#define LOCAL_DEBUG


/***** procedures for constraint selection *****/

/* Rates how difficult it's (probably) going to be to fulfill a constraint 
 * the more difficult, the higher the return value
 */
float compute_eq_constraint_difficulty(constr, limit) 
sem_elem_constr* constr;
float            limit;
{
   float  retval;

   switch (constraint_selection_strategy) {
   case 0: 
	 retval = 1.0;
	 break;

   case 1: 
	 retval = ((float) dag_tc_of_sem_elem_constr(constr, 
						     (int) limit));
	 break;

   case 2:
	 retval = (float) nr_of_top_level_constraints_satisfied_by_elem_constr(constr);
	 break;

   case 3:
	 retval = ((float) nr_of_top_level_constraints_satisfied_by_elem_constr(constr) 
		 * (float) dag_tc_of_sem_elem_constr(constr, (int) limit));
	 break; 

   default:
	 sam_error("compute_eq_constraint_difficulty: Unknown selection strategy", 
		   NULL, 2); 
   }

   return retval;
}



/* select a living (alive) constraint among the constraints of var,
   which has minimal difficulty or return the saved_equ_constr, if it
   is alive */
sem_elem_constr* select_constraint(var) 
WORD* var;
{
   sem_elem_constr*  bestConstraint = NULL;
   float             bestDifficulty = MAXFLOAT;
   sem_elem_constr*  constraint;
   sem_elem_constr*  savedEquConstr;
   float             difficulty;

   if (var == NULL)
      return NULL;

   savedEquConstr = ((constr_block*) GETPTR(*var))->saved_equ_constr;
   if (savedEquConstr  &&  is_sem_elem_constr_alive(savedEquConstr))
      return savedEquConstr;
 
   constraint = ((constr_block*) GETPTR(*var))->sem_equ_constr;

   /* process the eq_constraints of var */
   while (constraint) {

      /* is this constraint alive? */
      if (is_sem_elem_constr_alive(constraint)) {

	 difficulty = compute_eq_constraint_difficulty(constraint,
						       bestDifficulty); 
	 /* isn't it too difficult? */
	 if (difficulty < bestDifficulty) {
	    bestConstraint = constraint;
	    bestDifficulty = difficulty;
	 }

      } /* if */

      constraint = constraint->next;

   } /* while */

   return bestConstraint;
}



/* select a living (alive) constraint from the number newest
   constraints of semConstrList (if alive_flag is set, ccount only
   alive constraints): */ 
sem_constr_list  select_constr_from_list(semConstrList,
					 number,
					 alive_flag,
					 selectEasiest) 
sem_constr_list  semConstrList;
int              number;
int              alive_flag;
int              selectEasiest;
{
   sem_constr_list  semConstrListEle     = semConstrList;
   int              checkOnlyElemConstrs = FALSE;
   sem_constr_list  bestSemConstrListEle = NULL;
   int              bestApp;
   float            bestValue;
   int              curApp = 0;    /* initialization necessary because
				      curApp is only assigned for elem
				      constrs and only if
				      !selectEasiest */ 
   float            curValue;
   sem_child*       curChild;
   sem_junct*       curJunct;
   sem_elem_constr* curElemConstr;
   sem_junct*       parent;
   int              counter = 1;

   if (selectEasiest) {
      bestApp   = MAXNUMBER;
      bestValue = MAXFLOAT;
   }
   else {
      bestApp   = 0;
      bestValue = 0.0;
   }

   while (semConstrListEle  &&  counter <= number) {

      if (!checkOnlyElemConstrs
	  ||  semConstrListEle->child_type == CHILD_ELEM) {
	 
	 curChild = semConstrListEle->child;
	 
	 switch (semConstrListEle->child_type) {

	 case CHILD_ELEM:
	       curElemConstr = (sem_elem_constr*) curChild;
	       parent = curElemConstr->parent_junct;
	       switch (curElemConstr->type) {
	       case ELEM_EQU:
		     if (!selectEasiest) {
			curApp = nr_of_appearances(curElemConstr->var, 
						   curElemConstr->parent_junct);
		     }
		     curValue = compute_variable_difficulty(curElemConstr->var);
		     if (!checkOnlyElemConstrs) {
			checkOnlyElemConstrs = TRUE;
			bestSemConstrListEle = semConstrListEle;
			bestApp              = curApp;
			bestValue            = curValue;
			semConstrListEle     = semConstrListEle->next;
			counter++;
			continue;
		     }
		     break;
	       case DEACTIVATED:
		     semConstrListEle = semConstrListEle->next;
		     if (!alive_flag)
			counter++;
		     continue;
		     break;
	       default:
		     sam_error("select_constr_from_list: wrong elem constr type",
			       NULL, 2);
	       }
	       break;

	 case CHILD_JUNCT:
	       curJunct = (sem_junct*) curChild;
	       parent = curJunct->parent_junct;  /* == NULL, if
						    curJunct is a
						    child of
						    semantic_root */
	       switch (curJunct->type) {
	       case JUNCT_CONJ:
		     curValue = (float) curJunct->arity;
		     break;
	       case JUNCT_DISJ:
		     curValue = 1.0 / ((float) curJunct->arity);
		     break;
	       case DEACTIVATED:
		     semConstrListEle = semConstrListEle->next;
		     if (!alive_flag)
			counter++;
		     continue;
		     break;
	       default:
		     fprintf(stderr, "junctor = %d\n", curJunct->type);
		     sam_error("select_constr_from_list: wrong junctor type",
			       NULL, 2);
	       }
	       break;

	 default:
	       sam_error("select_constr_from_list: wrong child type",
			 NULL, 2);

	 }  /* switch (semConstrListEle->child_type) */
	 
	 if (    (select_oldest  &&  !parent)
	     ||  (selectEasiest  &&  curValue < bestValue)
	     ||  (!selectEasiest
		  &&  (curApp > bestApp
		       ||  (curApp == bestApp  &&  curValue > bestValue)))) {
	    bestSemConstrListEle = semConstrListEle;
	    bestApp              = curApp;
	    bestValue            = curValue;
	 }
      }  /* if */
      
      semConstrListEle = semConstrListEle->next;
      counter++;

   }  /* while */
   
   return bestSemConstrListEle;
}



/* select a living (alive) elementary constraint from the number
   newest constraint trees (if alive_flag is set, count only alive
   constraint trees): */
sem_elem_constr*  select_constr_from_tree(number, alive_flag)
int  number;
int  alive_flag;
{
   sem_constr_list   selectedListEle;
   sem_junct*        junct;

   selectedListEle = select_constr_from_list(semantic_root->sub_constr_list,
					     number,
					     alive_flag,
					     FALSE);

   while (selectedListEle  &&
	  selectedListEle->child_type != CHILD_ELEM) {
      junct = (sem_junct*) selectedListEle->child;
      switch (junct->type) {
      case JUNCT_CONJ:
	    selectedListEle = select_constr_from_list(junct->sub_constr_list,
						      MAXNUMBER,
						      alive_flag,
						      FALSE);
	    break;
      case JUNCT_DISJ:
	    selectedListEle = select_constr_from_list(junct->sub_constr_list,
						      MAXNUMBER,
						      alive_flag,
						      TRUE);
	    break;
      default:
	    sam_error("select_constr_from_tree: wrong junctor type", 
		      NULL, 2);
      }
   }

   if (selectedListEle)
      return ((sem_elem_constr*) selectedListEle->child);
   else
      return NULL;
}



/***** procedures for variable selection *****/

float compute_variable_difficulty(myVar)
/* Rates how difficult it's (probably) going to be to find an
   instantiation for myVar  
 * the more difficult, the higher the return value
 */
WORD* myVar;
{
  sem_elem_constr*  myConstr;
  float             retval;

  switch (variable_selection_strategy) {
     
  case 0:
	/* if myVar owns a living (alive) elem constr <==> return 1.0
	   otherwise                                  <==> return 0.0 */
	myConstr = ((constr_block*) GETPTR(*myVar))->sem_equ_constr;
	while (myConstr) {
	   if (is_sem_elem_constr_alive(myConstr))
	      return 1.0;
	   myConstr = myConstr->next;
	}
	retval = 0.0;
	break;

  case 1:
	retval = ((float) nr_of_constraints_at_variable(myVar));
	break;

  case 2: 
	retval = ((float) nr_of_top_level_constraints_at_variable(myVar));
	break;

  case 3:
	retval = ((float) nr_of_top_level_constraints_at_variable(myVar)
		  + AverageDagTC_of_SemEqConstrsAtVariable(myVar));
	break;

  case 4:
	retval = ((float) nr_of_top_level_constraints_at_variable(myVar)
		  + (0.1 * (float) nr_of_constraints_at_variable(myVar)));
	break;

  case 5:
	retval = ((float) max_nr_of_appearances_in_same_parent_AND_node(myVar));
	break;
	
  default:
	sam_error("compute_variable_difficulty: Unknown selection strategy", 
		  NULL, 2);

  } /* switch */

  return retval;
}



WORD* select_variable(varBlock, nrOfVars)
/* Select a variable from varBlock (which comprises nrOfVars variables)
 * by computing compute_variable_difficulty for each variable and
 * picking the most "difficult" one
 */
WORD* varBlock;
int   nrOfVars;
{
  WORD* bestVar = NULL;
  WORD* curVar;
  float bestValue = 0.0;
  float curValue;
  int   i; 
  
  /* Check all available variables */
  for(i = 0; i < nrOfVars; i++) {
    
    curVar = GETPTR(*(varBlock + i));
    
    if (GETTAG(*curVar) == T_FVAR) { /* Variable is still free */
      
      /* if curVar has equality constraints */
      if (((constr_block*)GETPTR(*curVar))->sem_equ_constr != NULL) {
	
	curValue = compute_variable_difficulty(curVar);
	
	/* remember the most "difficult" variable */
	if (curValue > bestValue) {
	  bestVar = curVar; 
	  bestValue = curValue;
	}
	
      } /* if curVar has equality constraints */
    } /* if variable is still free */
  } /* for */

   return bestVar;
}



/**************** procedures for backtracking ********************/

history_list  add_to_history(constraint, cons_fail_occ_flag,
			     trailPtr, cstackPtr, cCounterSp, heapPtr,
			     myHistoryPtr)
/* Create a new history-list-element and insert it as first element
 * into the list pointed to by myHistoryPtr
 */
sem_elem_constr*  constraint;
int               cons_fail_occ_flag;
s_trail*          trailPtr;
WORD*             cstackPtr;
c_counter*        cCounterSp;
WORD*             heapPtr;
history_list*     myHistoryPtr;
{
   history_list newList = (history_list) MALLOC(sizeof(history_list_ele));
   
   newList->constraint    = constraint;
   newList->cons_fail_occ = cons_fail_occ_flag;
   newList->trail_ptr     = trailPtr;
   newList->cstack_ptr    = cstackPtr;
   newList->c_counter_sp  = cCounterSp;
   newList->heap_ptr      = heapPtr;
   if (*myHistoryPtr == NULL)
      newList->ele_nr     = 1;
   else
      newList->ele_nr     = (*myHistoryPtr)->ele_nr + 1;
   newList->previous      = *myHistoryPtr;
   
   return newList;
}



/* Delete all subconstraints that have already been (implicitly)
   examined: 
   E.g. if a binding X leads to the fulfillment of the constraint X==t
   and is then being undone, all elementary (equality) constraints
   X==t can be deleted 
   returns FALSE, if root node has been deleted, i.e. the constraint net
                                                      is inconsistent 
   returns TRUE  otherwise
*/
int mark_as_deleted(myConstr)
sem_elem_constr* myConstr;
{
  sem_elem_constr*  curConstr;

  if (myConstr == NULL) 
     return FALSE;

  curConstr = ((constr_block*) GETPTR(*(myConstr->var)))->sem_equ_constr;

  while (curConstr) {
     if (is_sem_elem_constr_alive(curConstr)) {
	if (chk_arg(curConstr->right_side, bp, myConstr->right_side, bp)) {
	   if (delete_sem_elem_constr(curConstr) == 0)
	      return FALSE;
	}
     }
     curConstr = curConstr->next;
  }

  return TRUE;
}



/* Backtrack to previous history element */
int  backtrack(historyListPtr, choicepoint, historyForCanlListPtr)
history_list*  historyListPtr;
choice_pt*     choicepoint;
history_list*  historyForCanlListPtr;
{
   int               consistentDeleted = FALSE;
   history_list      tmpHistory;
   c_counter*        syn_junct = NULL;

   if (solution_computation_mode == NO_SOL_CMP)
      sam_error("wrong solution computation mode", NULL, 2);
   
   while (*historyListPtr != NULL  &&  !consistentDeleted) {

      untrail((*historyListPtr)->trail_ptr);
      tr                                     = (*historyListPtr)->trail_ptr;
      c_sp                                   = (*historyListPtr)->cstack_ptr;
      c_counter_sp                           = (*historyListPtr)->c_counter_sp;
      hp                                     = (*historyListPtr)->heap_ptr;
      choicepoint->consistency_fail_occurred = (*historyListPtr)->cons_fail_occ;
      
      if ((intern_cons_anl
	   ||  (cons_anl  &&  (*historyListPtr)->ele_nr == 1))
	  &&  solution_computation_mode != INST_TREES) {
	 pu_tr = pu_trail;  /* reset pu_trail */
	 if (un_protoco((*historyListPtr)->constraint->var, bp, 
			(*historyListPtr)->constraint->right_side, bp,
			1) != TRUE) {
	    sam_error("internal error in backtracking", NULL, 2);
	 }
	 if (pu_tr == pu_trail) {
	    /* c_immed_icanl_fails++;  */
	    sam_error("internal error in backtracking", NULL, 2);
	 }
	 else {
	    /* lcanl/gcanl werden wie icanl behandelt */
	    syn_junct = c_gencounter(I_CONS_ANL, choicepoint, 0);
	    c_gen_optimize(syn_junct);
	    gen_icanl_constrs++;
	 }
      }

      consistentDeleted = mark_as_deleted((*historyListPtr)->constraint);
      
      if (cons_anl  &&  solution_computation_mode != INST_TREES) {

	 /* remove all elements from the list
	    (*historyForCanlListPtr) which have a smaller ele_nr than
	    the current history element: */
	 while (*historyForCanlListPtr  &&  
		(*historyForCanlListPtr)->ele_nr > (*historyListPtr)->ele_nr) {
	    tmpHistory = *historyForCanlListPtr;
	    *historyForCanlListPtr = (*historyForCanlListPtr)->previous;
	    FREE(tmpHistory, sizeof(history_list_ele));
	 }

	 if ((*historyListPtr)->ele_nr == 1) {
	    /* remove current history element: */
	    tmpHistory = *historyListPtr;
	    *historyListPtr = (*historyListPtr)->previous;
	    FREE(tmpHistory, sizeof(history_list_ele));
	 }
	 else {
	    /* remove current history element from the list
	       (*historyListPtr) and insert it into the list
	       (*historyForCanlListPtr): */
	    tmpHistory = *historyListPtr;
	    *historyListPtr = (*historyListPtr)->previous;
	    tmpHistory->previous = *historyForCanlListPtr;
	    *historyForCanlListPtr = tmpHistory;
	 }

      }
      else {
	 /* remove current history element: */
	 tmpHistory = *historyListPtr;
	 *historyListPtr = (*historyListPtr)->previous;
	 FREE(tmpHistory, sizeof(history_list_ele));
      }
      
   } /* while */
   
   return consistentDeleted;
}



/****************** procedures for instantiation ********************/

/* The following macros are implicit arguments for the unification
   macros (like UN_FVAR) */ 
#define FAIL_U       {return FALSE;}
#define UNIF_FAIL(x) { \
switch (solution_computation_mode) { \
case NO_SOL_CMP: \
      sam_error("wrong solution computation mode", NULL, 2); \
case LOCAL_CHK:  loc_cons_unif_fails++; break; \
case GLOBAL_CHK: gl_cons_unif_fails++;  break; \
case INST_TREES: inst_anl_unif_fails++; break; \
default: \
      sam_error("undefined solution computation mode", NULL, 2); \
} \
FAIL_U}


int is_sem_equ_constr_fulfilled(myConstr)
/* Check if myConstr can still be fulfilled */
sem_elem_constr* myConstr;
{
   WORD* myVar;
   WORD* myRightSide;
   
   if (!myConstr)
      return FALSE;
   
   if (!is_sem_elem_constr_alive(myConstr)) {
      switch (myConstr->type) {
      case DEACTIVATED:
	    return TRUE;
      case DELETED:
	    return FALSE;
      default: 
	    sam_error("is_sem_equ_constr_fulfilled: wrong type of elem constr",
		      NULL,2);
      }
   }

   /* Unify right and left side to see if myConstr can still be fulfilled */
   myVar       = deref(myConstr->var, bp);
   myRightSide = deref(myConstr->right_side, bp);
   UNIFY(myVar, bp, myRightSide, bp);

   return TRUE;
}

#undef FAIL_U
#undef UNIF_FAIL



int  instantiate(nrTreesToCheck, alive_flag, historyPtr)
int            nrTreesToCheck;
int            alive_flag;
history_list*  historyPtr;
{
   sem_elem_constr* constraint = select_constr_from_tree(nrTreesToCheck,
							 alive_flag);
   int              savedConsFailOccFlag = chp->consistency_fail_occurred; 
   s_trail*         savedTr              = tr;
   WORD*            savedCSp             = c_sp;
   c_counter*       savedCCounterSp      = c_counter_sp;
   WORD*            savedHp              = hp;

   while (constraint) {

      if (is_sem_equ_constr_fulfilled(constraint)) {

	 /* Save variable, constraint and trail pointer: */ 
	 *historyPtr = add_to_history(constraint,
				      savedConsFailOccFlag,
				      savedTr, savedCSp,
				      savedCCounterSp, savedHp,
				      historyPtr);
      }
      else {

	 /* Backtrack recent variable and constraint choice: */
	 untrail(savedTr);
	 chp->consistency_fail_occurred = savedConsFailOccFlag;
	 tr                             = savedTr;
	 c_sp                           = savedCSp;
	 c_counter_sp                   = savedCCounterSp;
	 hp                             = savedHp;

	 if (mark_as_deleted(constraint) == FALSE) {
	    return FALSE;
	 }
	 
      }

      constraint = select_constr_from_tree(nrTreesToCheck,
					   alive_flag); 
      savedConsFailOccFlag = chp->consistency_fail_occurred;
      savedTr              = tr;
      savedCSp             = c_sp;
      savedCCounterSp      = c_counter_sp;
      savedHp              = hp;
   }
   
   return TRUE;
}



/*************** procedures for consistency checks *****************/

/* check the consistency of the nrTreesToCheck newest alive constraint
   trees.
   
   after the check:
   - untrail variable bindings made during check
   - set the consistency_fail_occurred flag in choicepoint
     if the constraint net is not consistent
   - remove the history list

   return TRUE  if the constraints are consistent
   return FALSE otherwise
   */
int are_sem_equ_constraints_consistent(nrTreesToCheck,
				       alive_flag,
				       choicepoint,
				       historyListPtr,
				       historyForCanlListPtr)
int            nrTreesToCheck;
int            alive_flag;
choice_pt*     choicepoint;
history_list*  historyListPtr;
history_list*  historyForCanlListPtr;
{
   int    successful_instantiated;
   int    successful_backtracked;

   do {
      /* try to instantiate the constraint trees: */
      successful_instantiated = instantiate(nrTreesToCheck,
					    alive_flag,
					    historyListPtr); 
      
      if (!successful_instantiated) {
	 /* Instantiation not possible ==> backtrack */
	 successful_backtracked = backtrack(historyListPtr, 
					    choicepoint,
					    historyForCanlListPtr);
      }
      
      /* continue until an instantiation has been found or no more
	 backtracking is possible: */ 
   } while (!successful_instantiated  &&  successful_backtracked);

   return successful_instantiated;
}



/* - check the global consistency of the actual constraint net
   - return TRUE   if the net is consistent
     return FALSE  otherwise
 */
int global_consistency(chpt, keep_instantiations)
choice_pt* chpt;
int        keep_instantiations;
{
   int               successful_instantiated;
   history_list      tmpHistory;
   history_list      history = NULL;
   history_list      historyForCanl = NULL;

#define DEFINE_untrail_after_global_consistency_check
#ifdef DEFINE_untrail_after_global_consistency_check
#warning untrailing after global_consistency_check !!!!!!!!!!!!!!!!!
   s_trail*    save_tr            = tr;
   WORD*       save_c_sp          = c_sp;
   c_counter*  save_c_counter_sp  = c_counter_sp;
   WORD*       save_hp            = hp;
   int         save_cons_fail_occ = chpt->consistency_fail_occurred;
   int         nrOfArgs = GETARITY(bp->lcbptr->ps_symb);
   int         i;
   
   printf("\n\n##########################################\n\n");
   if (nrOfArgs == 0) {
      printf("at head literal %s (clause %d) ",
	     GETPRINTNAME(bp->lcbptr->ps_symb),
	     GETNUMBER(bp->lcbptr->clause_nr));
      fflush(stdout);
   }
   else {
      printf("at head literal %s(",
	     GETPRINTNAME(bp->lcbptr->ps_symb));
      for ( i = 0 ; i < nrOfArgs ; i++ ) {
	 disp_(stdout,  
	       deref((code + GETVAL(bp->lcbptr->argv + i)), bp),
	       bp);
	 if (i < nrOfArgs - 1) {
	    printf(", ");
	 }
      }
      printf(") (clause %d) ", GETNUMBER(bp->lcbptr->clause_nr));
      printf("(total_inf = %d) ", total_inf);
      /*
      printf("\nwith semantic constraint ");
      print_sem_constr_tree(semantic_root, 0);
      printf("\n");
      */
      fflush(stdout);
   }
#endif

#ifdef STATISTICS1
   gl_cons_chks++;
   adapt_gl_cons_chks++;
#endif

   if (solution_computation_mode != NO_SOL_CMP)
      sam_error("Wrong solution computation mode.", NULL, 2);
   solution_computation_mode = GLOBAL_CHK;

   successful_instantiated = are_sem_equ_constraints_consistent(MAXNUMBER,
								TRUE,
								chpt,
								&history,
								&historyForCanl);

   if (solution_computation_mode != GLOBAL_CHK)
      sam_error("Wrong solution computation mode.", NULL, 2);
   solution_computation_mode = NO_SOL_CMP;

#ifdef DEFINE_untrail_after_global_consistency_check
   printf("global consistency check returns %d\n",
	  successful_instantiated); 
#endif
   if (keep_instantiations) {

      /* DEBUGGING */
      printf("\n###############solution computed at default query###############\n");
      fflush(stdout);

      if (!successful_instantiated) {
	 print_sem_constr_tree(semantic_root, 0);
	 if (history)
	    printf("\nnr if history elements to backtrack: %d\n",
		   history->ele_nr);
	 else
	    printf("\nno more history elements to backtrack\n");
	 fflush(stdout);
	 sam_error("global_consistency: gl cons chk failed at default query",
		   NULL, 2);
      }
      else {
	 return TRUE;
      }
   }
   
   /* untrail variable bindings, clean constraint stack and heap, and
      remove elements of history list: */
   if (history) {
      while (history->previous) {
	 tmpHistory = history;
	 history = history->previous;
	 FREE(tmpHistory, sizeof(history_list_ele));
      }
      untrail(history->trail_ptr);
      tr           = history->trail_ptr;
      c_sp         = history->cstack_ptr;
      c_counter_sp = history->c_counter_sp;
      hp           = history->heap_ptr;
      chpt->consistency_fail_occurred = history->cons_fail_occ;
      FREE(history, sizeof(history_list_ele));
   }

#ifdef DEFINE_untrail_after_global_consistency_check
   untrail(save_tr);
   tr                              = save_tr;
   c_sp                            = save_c_sp;
   c_counter_sp                    = save_c_counter_sp;
   hp                              = save_hp;
   chpt->consistency_fail_occurred = save_cons_fail_occ;
#endif

   if (!successful_instantiated) {
      chpt->consistency_fail_occurred = TRUE;
#ifdef STATISTICS1
      gl_cons_fails++;
      adapt_gl_cons_fails++;
#endif
   }

   return successful_instantiated;
}



/* - check the local consistency of the actual constraint net
   - return TRUE   if the net is local consistent
     return FALSE  otherwise
 */
int try_to_instantiate_trees(chpt, nrTreesToCheck)
choice_pt*  chpt;
int         nrTreesToCheck;
{
   int               successful_instantiated = TRUE;
   history_list      tmpHistory;
   history_list      history = NULL;
   history_list      historyForCanl = NULL;

#define DEFINE_untrail_after_try_to_instantiate_trees
#ifdef DEFINE_untrail_after_try_to_instantiate_trees
#warning untrailing after try_to_instantiate_trees !!!!!!!!!!!!!!!!!
   s_trail*    save_tr            = tr;
   WORD*       save_c_sp          = c_sp;
   c_counter*  save_c_counter_sp  = c_counter_sp;
   WORD*       save_hp            = hp;
   int         save_cons_fail_occ = chpt->consistency_fail_occurred;
   int         nrOfArgs = GETARITY(bp->lcbptr->ps_symb);
   int         i;
   
   if (nrOfArgs == 0) {
      printf("at head literal %s (clause %d) ",
	     GETPRINTNAME(bp->lcbptr->ps_symb),
	     GETNUMBER(bp->lcbptr->clause_nr));
      fflush(stdout);
   }
   else {
      printf("at head literal %s(",
	     GETPRINTNAME(bp->lcbptr->ps_symb));
      for ( i = 0 ; i < nrOfArgs ; i++ ) {
	 disp_(stdout, 
	       deref((code + GETVAL(bp->lcbptr->argv + i)), bp), 
	       bp);
	 if (i < nrOfArgs - 1) {
	    printf(", ");
	 }
      }
      printf(") (clause %d) ", GETNUMBER(bp->lcbptr->clause_nr));
      printf("(total_inf = %d) ", total_inf);
      if (total_inf == 1071) {
	 printf("\nwith semantic constraint ");
	 print_sem_constr_tree(semantic_root, 0);
	 printf("\n");
	 print_syn_constrs_of_chpt(chpt);
      }
      fflush(stdout);
   }
#endif

   /* compute instantiations for the nrTreesToCheck newest
      constraint trees: */
   successful_instantiated = are_sem_equ_constraints_consistent(nrTreesToCheck,
								TRUE,
								chpt,
								&history,
								&historyForCanl);

#ifdef DEFINE_untrail_after_try_to_instantiate_trees
   printf("local consistency check returns %d\n",
	  successful_instantiated); 
#endif

   if (successful_instantiated  &&  cons_anl  &&  historyForCanl) {

      if ((historyForCanl  &&  !history)
	  ||  (historyForCanl  
	       &&  history
	       &&  historyForCanl->ele_nr > history->ele_nr + 1))
	 sam_error("error concerning historyForCanl", NULL, 2);

      /*
      if (historyForCanl) {
	 printf("\n\nAFTER LOC-CONS-CHK:\nprinting historyForCanl:\n");
	 printf("========================\n");
	 fflush(stdout);
      }
      tmpHistory = historyForCanl;
      while (tmpHistory) {
	 printf("ele_nr = %d\n", tmpHistory->ele_nr);
	 fflush(stdout);
	 tmpHistory = tmpHistory->previous;
      }
      */

      /* generate lcanl from the historyForCanl list and the history
	 list: */
      while (historyForCanl) {
	 /* up to the pointers stored in the current element of
	    historyForCanl, untrail variable bindings, clean
	    constraint stack and heap: */
	 untrail(historyForCanl->trail_ptr);
	 tr           = historyForCanl->trail_ptr;
	 c_sp         = historyForCanl->cstack_ptr;
	 c_counter_sp = historyForCanl->c_counter_sp;
	 hp           = historyForCanl->heap_ptr;
	 chpt->consistency_fail_occurred = historyForCanl->cons_fail_occ;

	 /* except for the (historyForCanl->ele_nr - 1) last elements,
	    remove list: */
	 while (history->ele_nr >= historyForCanl->ele_nr) {
	    tmpHistory = history;
	    history = history->previous;
	    FREE(tmpHistory, sizeof(history_list_ele));
	 }

	 /* generate lcanl: */
	 chpt->lcanl_list = generate_canl(historyForCanl, 
					  history,
					  chpt->lcanl_list);
#ifdef STATISTICS1
	 lcanlcount++;
#endif
	 
	 /* remove current element of the historyForCanl list: */
	 tmpHistory = historyForCanl;
	 historyForCanl = historyForCanl->previous;
	 FREE(tmpHistory, sizeof(history_list_ele));
      }

      /* untrail remaining variable bindings, clean constraint stack
	 and heap, and remove remaining elements of history list: */
      if (history) {
	 while (history->previous) {
	    tmpHistory = history;
	    history = history->previous;
	    FREE(tmpHistory, sizeof(history_list_ele));
	 }
	 untrail(history->trail_ptr);
	 tr           = history->trail_ptr;
	 c_sp         = history->cstack_ptr;
	 c_counter_sp = history->c_counter_sp;
	 hp           = history->heap_ptr;
	 chpt->consistency_fail_occurred = history->cons_fail_occ;
	 FREE(history, sizeof(history_list_ele));
      }
#ifdef DEFINE_untrail_after_try_to_instantiate_trees
      untrail(save_tr);
      tr                              = save_tr;
      c_sp                            = save_c_sp;
      c_counter_sp                    = save_c_counter_sp;
      hp                              = save_hp;
      chpt->consistency_fail_occurred = save_cons_fail_occ;
#endif
      
      /*
      {
	 a_list*   lcanl_list_ptr;

	 printf("\n\nprinting canl list:\n");
	 printf("===================\n");
	 fflush(stdout);
	 lcanl_list_ptr = chpt->lcanl_list;
	 while (lcanl_list_ptr) {
	    printf("\nleft side:\n");
	    print_instantiations(lcanl_list_ptr->left, 
				 lcanl_list_ptr->arity, 
				 TRUE);
	    printf("right side:\n");
	    print_instantiations(lcanl_list_ptr->right, 
				 lcanl_list_ptr->arity, 
				 FALSE);
	    fflush(stdout);
	    lcanl_list_ptr = lcanl_list_ptr->next;
	 }
      }
      */

      /* generate lcanl constraints: */
      if (!anl_gen_constr_from_list(chpt->lcanl_list, 
				    chpt, 
				    LOC_CONS_ANL)) {
	 sam_error("immediate lcanl-constr fail", NULL, 2);
      }
      
   }
   else {
      /* untrail variable bindings, clean constraint stack and heap, and
	 remove elements of history list: */
      if (history) {
	 while (history->previous) {
	    tmpHistory = history;
	    history = history->previous;
	    FREE(tmpHistory, sizeof(history_list_ele));
	 }
	 untrail(history->trail_ptr);
	 tr           = history->trail_ptr;
	 c_sp         = history->cstack_ptr;
	 c_counter_sp = history->c_counter_sp;
	 hp           = history->heap_ptr;
	 chpt->consistency_fail_occurred = history->cons_fail_occ;
	 FREE(history, sizeof(history_list_ele));
      }
#ifdef DEFINE_untrail_after_try_to_instantiate_trees
      untrail(save_tr);
      tr                              = save_tr;
      c_sp                            = save_c_sp;
      c_counter_sp                    = save_c_counter_sp;
      hp                              = save_hp;
      chpt->consistency_fail_occurred = save_cons_fail_occ;
#endif
   }
   
   return (successful_instantiated);
}



int one_consistency_check(varList, nrVars, GETPTR_flag, deleted_flag)
/* The constraints of a variable are one-consistent, if there's an
 * instantiation for this variable which fulfills these constraints.
 * return TRUE, if var's constraints are one-consistent
 * return FALSE otherwise */
WORD*  varList;
int    nrVars;
int    GETPTR_flag;
int    *deleted_flag;
{
   s_trail*          savedTrail;
   WORD*             variable;
   sem_elem_constr*  constraint;
   int               counter = 0;
   int               varIndex = 0;
#define DEFINE_untrail_after_one_consistency_check
#ifdef DEFINE_untrail_after_one_consistency_check
#warning untrailing after one_consistency_check !!!!!!!!!!!!!!!!!
   s_trail*          savedTrail2 = tr;
#endif

   *deleted_flag = FALSE;

   while (counter < nrVars) {
      
      if (GETPTR_flag)
	 variable = deref(GETPTR(*(varList + (varIndex % nrVars))), bp);
      else
	 variable = deref(varList + (varIndex % nrVars), bp);

      if (GETTAG(*variable) != T_FVAR) {
	 counter++;
	 varIndex++;
	 continue;   /* variable is bound to a non-variable term */
      }
      
      savedTrail = tr;
      constraint = ((constr_block*) GETPTR(*variable))->sem_equ_constr;
      while (constraint) {
	 
	 if (is_sem_elem_constr_alive(constraint)) {
	    
	    if (is_sem_equ_constr_fulfilled(constraint)) {
	       /* the right_side of constraint is a possible instantiation
		  for variable => clean up and finish */
	       untrail(savedTrail);
	       tr = savedTrail;
	       if (!one_consistency_elimination)
		  break;
	    }
	    else {
	       /* the right_side of constraint is NOT a possible
		  instatiation for variable */ 
	       untrail(savedTrail);
	       tr = savedTrail;
	       if (!mark_as_deleted(constraint)) {
		  /* root deleted => constraint net is inconsistent */
#ifdef DEFINE_untrail_after_one_consistency_check
		  untrail(savedTrail2);
		  tr = savedTrail2;
#endif
		  return FALSE;
	       }
	       else {
		  *deleted_flag = TRUE;
		  if (GETTAG(*variable) != T_FVAR) {
		     /* variable has been bound during mark_as_deleted */
		     break;
		  }
		  else {
		     savedTrail = tr;
		     counter = -1;
		  }
	       }
	    }
	    
	 } /* if constraint is alive */
	 
	 constraint = constraint->next;
	 
      } /* while (constraint) */

      counter++;
      varIndex++;

   } /* while (counter < nrVars) */

#ifdef DEFINE_untrail_after_one_consistency_check
   untrail(savedTrail2);
   tr = savedTrail2;
#endif

   return TRUE;
}
