/******************************************************/
/*    S E T H E O                                     */
/*                                                    */
/* FILE: sem_consistency.c                            */
/* VERSION:                                           */
/* DATE:                                              */
/* AUTHOR: Rauschmayer & Sacher                       */
/* NAME OF FILE:                                      */
/* DESCR:                                             */
/* MOD:                                               */
/* BUGS:                                              */
/******************************************************/

#include <stdio.h>
#include <math.h>
#include "tags.h"
#include "machine.h"
#include "symbtab.h"
#include "deref.h"
#include "errors.h"
#include "sem_constr.h"
#include "sem_consistency.h"
#include "termcomplexity.h"
#include "unification.h"
#include "unifeq.h"
#include "unitlemma.h"

/*************** constants ******************/

#define LOCAL_DEBUG



/***** procedures for constraint evaluation and selection *****/

int  dag_tc_of_sem_elem_constr(constr, limit)
   sem_elem_constr*  constr;
   int               limit;
{
   s_trail*  saved_tr = tr;
   int       retval = dag_termcomplexity(constr->right_side, 1, bp,
					 limit); 
   untrail(saved_tr);
   tr = saved_tr;

   return retval;
}



/* Rates how difficult it's (probably) going to be to fulfill a constraint 
 * the more difficult, the higher the return value
 */
float compute_eq_constraint_difficulty(constr, limit) 
sem_elem_constr* constr;
float            limit;
{
   switch (1) {/* (eq_constraint_selection_strategy) {*/
   case 0: 
	 return 1.0;
   case 1: 
	 return ((float) dag_tc_of_sem_elem_constr(constr, 
						   (int) limit)); 
   default:
	 sam_error("compute_eq_constraint_difficulty: Unknown selection strategy", 
		   NULL, 2); 
   } 
}



/* select a living (alive) constraint among the constraints of var,
   which has minimal difficulty */ 
sem_elem_constr* select_constraint(var) 
WORD* var;
{
   sem_elem_constr*  bestConstraint = NULL;
   float             bestDifficulty = MAXFLOAT;
   sem_elem_constr*  constraint;
   float             difficulty;

   if (var==NULL)
      return NULL;

   constraint = ((constr_block *)GETPTR(*var))->sem_equ_constr;

   /* process the eq_constraints of var */
   while(constraint) {

      /* is this constraint alive? */
      if (is_sem_elem_constr_alive(constraint)) {

	 difficulty = compute_eq_constraint_difficulty(constraint,
						       bestDifficulty); 
	 /* isn't it to difficult? */
	 if (difficulty < bestDifficulty) {
	    bestConstraint = constraint;
	    bestDifficulty = difficulty;
	 }

      } /* if */

      constraint = constraint->next;

   } /* while */

   return bestConstraint;
}



/***** procedures for variable evaluation and selection *****/

int  NrTopLevelConstraintsAtVariable(var)
   WORD*  var;
{
   sem_constr_list   root_list, root_list_ele;
   int               length;

   if (GETTAG(*var) != T_FVAR)
      sam_error("NrTopLevelConstraintsAtVariable: Variable is already bound.",
		NULL, 2);
   
   root_list = compute_list_of_root_nodes_at_variable(var);

   /* Count elements of root_list and free root_list: */
   length = 0;
   while (root_list != NULL) {
      length++;
      root_list_ele = root_list;
      root_list = root_list->next;
      FREE(root_list_ele, sizeof(sem_constr_list_ele));
   }

   return length;
}



float  AverageDagTC_of_SemEqConstrsAtVariable(var)
   WORD*  var;
{
   sem_elem_constr*  constr; 
   int               nr_of_sem_elem_constrs = 0;
   int               sum_of_dag_tcs         = 0;
   
   if (GETTAG(*var) != T_FVAR)
      sam_error("AverageDagTC_of_SemEqConstrsAtVariable: Variable is already bound.",
		NULL, 2);
   
   constr = ((constr_block*) GETPTR(*var))->sem_equ_constr; 

   if (constr == NULL)
      return 0.0;

   while (constr != NULL) {
      nr_of_sem_elem_constrs++;
      sum_of_dag_tcs += dag_tc_of_sem_elem_constr(constr, MAXNUMBER);
      constr = constr->next;
   }
   
   return ((float) sum_of_dag_tcs / (float) nr_of_sem_elem_constrs);
}



float compute_variable_difficulty(myVar)
/* Rates how difficult it's (probably) going to be to find an instantiation for myVar 
 * the more difficult, the higher the return value
 */
WORD* myVar;
{
  int              count = 0;  /* nr of living (alive) elem constrs at myVar */
  sem_elem_constr* myConstr;

  switch (1) { /* (variable_selection_strategy) { */
  case 0:
    myConstr = ((constr_block*) GETPTR(*myVar))->sem_equ_constr;
    while (myConstr) {
      if (is_sem_elem_constr_alive(myConstr))
	 count++;
      myConstr = myConstr->next;
    }
    if (count)
       return 1.0;
    else
       return -1.0;
  case 1:
    /* Count the number of myVar's living (alive) subconstraints (in count) */
    myConstr = ( (constr_block*)GETPTR(*myVar) )->sem_equ_constr;
    while (myConstr) {
      if (is_sem_elem_constr_alive(myConstr))
	 count++;
      myConstr = myConstr->next;
    }
    if (count)
       return ((float) 1) / ((float) count);
    else
       return -1.0;
  default:
    sam_error("compute_variable_difficulty: Unknown selection strategy", NULL, 2);
  }
}




WORD* select_variable(varBlock, nrOfVars)
/* Select a variable from varBlock (which comprises nrOfVars variables)
 * by computing compute_variable_difficulty for each variable and
 * picking the most "difficult" one
 */
WORD* varBlock;
int   nrOfVars;
{
   WORD* bestVar = NULL;
   WORD* curVar;
   float bestValue = -1.0;
   float curValue;
   int   i; 
   
   for(i = 0; i < nrOfVars; i++) {
      /* Check all available variables */
      
      if (GETTAG(*(varBlock + i)) != T_MARKED_FVAR) {
	 /* variable only has to be checked if it`s not been marked yet */
	 
	 curVar = deref(GETPTR(*(varBlock + i)), bp);
	 
	 if (GETTAG(*curVar) == T_FVAR) { /* Variable is still free */
	    
	    /* if curVar has equality constraints */
	    if (((constr_block*)GETPTR(*curVar))->sem_equ_constr != NULL) {
	       
	       curValue = compute_variable_difficulty(curVar);

	       if (curValue > bestValue) {
                  /* remember the most "difficult" variable */
		  bestVar = curVar; 
		  bestValue = curValue;
	       }
	       
	    } /* if curVar has equality constraints */
	 }
	 else { /* Variable is bound, mark it */
	    trail_var(varBlock + i);
	    GENOBJ(*(varBlock + i), 0, T_MARKED_FVAR);
	 } /* if */

      } /* if */

   }
   
   return bestVar;
}



/**************** procedures for backtracking ********************/

history_list  add_to_history(var, constraint, trailPtr, myHistoryPtr)
/* Create a new history-list-element and insert it as first element
 * into the list pointed to by myHistoryPtr
 */
WORD*             var;
sem_elem_constr*  constraint;
s_trail*          trailPtr;
history_list*     myHistoryPtr;
{
   history_list newList = (history_list) MALLOC(sizeof(history_list_ele));
   
   newList->variable    = var;
   newList->constraint  = constraint;
   newList->trail_ptr   = trailPtr;
   newList->previous    = *myHistoryPtr;
   
   return newList;
}



/* Delete all subconstraints that have already been (implicitly)
   examined: 
   E.g. if a binding X leads to the fulfillment of the constraint X==t
   and is then being undone, all elementary (equality) constraints
   X==t can be deleted 
*/
int mark_as_deleted(myConstr)
sem_elem_constr* myConstr;
{
  sem_elem_constr*  curConstr;

  if (myConstr == NULL) 
     return FALSE;

  curConstr = ((constr_block*) GETPTR(*(myConstr->var)))->sem_equ_constr;

  while (curConstr) {
     if (is_sem_elem_constr_alive(curConstr)) {
	if (chk_arg(curConstr->right_side, bp, myConstr->right_side, bp)) {
	   if (delete_sem_elem_constr(curConstr) == 0)
	      return FALSE;
	}
     }
     curConstr = curConstr->next;
  }

  return TRUE;
}




/* Backtrack to previous history element */
history_list  backtrack_to_prev_history_ele(history)
history_list  history;
{
   int           consistentMarked = FALSE;
   history_list  saveHistory;
   
   while (history != NULL  &&  !consistentMarked) {
      
      untrail(history->trail_ptr);
      tr = history->trail_ptr;

      consistentMarked = mark_as_deleted(history->constraint);
   
      saveHistory = history;
      history = history->previous;
      FREE(saveHistory, sizeof(history_list_ele));
   }

   return history;
}



/****************** procedures for instantiation ********************/

/* The following macros are implicit arguments for the unification
   macros (like UN_FVAR) */ 
#define FAIL_U       return  FALSE;
#define UNIF_FAIL(x) FAIL_U

int is_sem_equ_constr_fulfilled(myConstr)
/* Check if myConstr can still be fulfilled */
sem_elem_constr* myConstr;
{
  WORD* temp;

  if (myConstr) {
    temp =  deref(myConstr->right_side, bp);
    /* Unify right and left side to see if myConstr can still be fulfilled */
    UNIFY(myConstr->var, bp, temp, bp);
    return TRUE;
  }
  else
    return FALSE;
}



int  instantiate(varBlock, nrOfVars, historyPtr)
WORD*          varBlock;
int            nrOfVars;
history_list*  historyPtr;
{
   WORD*             var = select_variable(varBlock, nrOfVars);
   sem_elem_constr*  constraint = select_constraint(var);
   s_trail*          savedTr = tr;

   while (var!=NULL) {
	
      if (is_sem_equ_constr_fulfilled(constraint)) {
	 /* Save variable, constraint and trail pointer: */ 
	 *historyPtr = add_to_history(var, 
				      constraint, 
				      savedTr,
				      historyPtr);
      }
      else {
	 /* Backtrack recent variable and constraint choice: */
	 untrail(savedTr);
	 tr = savedTr;

	 if (mark_as_deleted(constraint) == FALSE)
	    return FALSE;
      }

      /* to avoid variable alternation insert the following condition:
	 if (GETTAG(*var) != T_FVAR) */
      var = select_variable(varBlock, nrOfVars);
      constraint = select_constraint(var);
      savedTr = tr;

   }
   
   return TRUE;
}



/*************** procedures for consistency checks *****************/

int are_sem_equ_constraints_consistent(varBlock, nrOfVars)
/* check the consistency of the equality constraints of the free
   variables in varBlock 
   return TRUE  if the constraints are consistent
   return FALSE otherwise
   */
WORD* varBlock;
int   nrOfVars;
{
   int              successful = TRUE;
   history_list     history    = NULL;
   history_list     saveHistory;

   do { 
      successful = instantiate(varBlock, nrOfVars, &history);
      
      if (!successful)
	 history = backtrack_to_prev_history_ele(history);
      
   } while (!successful  &&  history!=NULL);
   
   while (history) {
      saveHistory = history;
      history = history->previous;
      FREE(saveHistory, sizeof(history_list_ele));
   }

   return successful;
}



int global_consistency(curChp, compute_solution)
/* check the consistency of the actual constraint net
 * return TRUE, if the net is consistent
 * retrun FALSE  otherwise
 */
choice_pt* curChp;
int        compute_solution;
{
   s_trail* savedTrail = tr;
   int      retVal;   /* SEM CONSTRS TIGHTENED */
 
   /* SEM CONSTRS TIGHTENED */
   checking_global_consistency = TRUE;
   retVal = are_sem_equ_constraints_consistent(curChp->block_fvars,
					       curChp->nr_fvars); 
   checking_global_consistency = FALSE;
   
   if (retVal == TRUE  &&  !compute_solution) {
      untrail(savedTrail);
      tr = savedTrail;
   }

   /* CONSISTENCY FAIL OCCURRED */
   if (retVal == FALSE)
      curChp->consistency_fail_occurred = TRUE;

   return retVal;
}
