/******************************************************/
/*    S E T H E O                                     */
/*                                                    */
/* FILE: sem_constr.c                                 */
/* VERSION:                                           */
/* DATE:                                              */
/* AUTHOR: Rauschmayer & Sacher                       */
/* NAME OF FILE:                                      */
/* DESCR:                                             */
/* MOD:                                               */
/* BUGS:                                              */
/******************************************************/

#include <stdio.h>
#include "tags.h"
#include "machine.h"
#include "deref.h"
#include "errors.h"
#include "disp.h"
#include "unitlemma.h"
#include "sem_constr.h"
#include "univ_constr.h"
#include "unification.h"



/*************** constants ******************/

#define LOCAL_DEBUG



/*************** procedures for tree output ******************/

void print_indentation(indent)
int indent;
{
  printf("\n");
  while(indent > 0) {
    printf("    ");
    indent --;
  }
}



void print_sem_elem_constr(self, indent)
sem_elem_constr* self;
int              indent;
{
  sem_elem_constr* curConstr = NULL;

  switch(self->type) {
  case ELEM_EQU:
    print_indentation(indent);
    disp_(stdout, self->var, bp); 
    printf(" == ");
    fflush(stdout);
    curConstr = ( (constr_block*)GETPTR(*(self->var)) )->sem_equ_constr;
    break;
  case ELEM_DIFF:
    print_indentation(indent);
    disp_(stdout, self->var, bp); 
    printf(" =/= ");
    fflush(stdout);
    curConstr = ( (constr_block*)GETPTR(*(self->var)) )->sem_diff_constr;
    break;
  case DELETED:
    /* in this case this part of the tree shouldn't be displayed */
    print_indentation(indent);
    printf("DELETED");
    fflush(stdout);
    return;
  case DEACTIVATED:
    /* in this case this part of the tree shouldn't be displayed */
    print_indentation(indent);
    printf("DEACTIVATED");
    fflush(stdout);
    return;
  default:
    sam_error("Wrong type in elementary subconstraint.",NULL,0);
  }

  disp_(stdout, self->right_side, bp); 
  printf(" (self: "GETVALFORMAT", parent: "GETVALFORMAT", next: "GETVALFORMAT")", 
	 (WORD_cast) self, 
	 (WORD_cast) self->parent_junct, 
	 (WORD_cast) self->next); 
  fflush(stdout);
  
  /* Does the variable have the correct back-pointer to the constraint? */
  if (GETTAG(*(self->var)) != T_FVAR) {
     sam_error("print_sem_elem_constr: back-pointer to bound variable.",
	       NULL,0);
     return;
  }
  while(curConstr && (curConstr != self)) {
    curConstr = curConstr->next;
  }
  if (!curConstr)
     sam_error("print_sem_elem_constr: NO BACK-POINTER!",NULL,0);
   
}



void print_sem_constr_tree(self, indent)
/* Display junctor self and its sons with the indentation indent */
sem_junct* self;
int        indent;
{
  sem_constr_list curEle = self->sub_constr_list;

  switch(self->type) {
  case JUNCT_DISJ:
    print_indentation(indent);
    printf("OR");
    fflush(stdout);
    break;
  case JUNCT_CONJ:
    print_indentation(indent);
    printf("AND");
    fflush(stdout);
    break;
  case DELETED:
    /* in this case this part of the tree shouldn't be displayed */
    print_indentation(indent);
    printf("DELETED");
    fflush(stdout);
    return;
  case DEACTIVATED:
    /* in this case this part of the tree shouldn't be displayed */
    print_indentation(indent);
    printf("DEACTIVATED");
    fflush(stdout);
    return;
  default:
    printf("Wrong junctor type: %d",self->type);
  }
  printf(" (self: "GETVALFORMAT", type: %i, arity: %i, parent: "GETVALFORMAT")",
	 (WORD_cast) self,
	 self->type,
	 self->arity,
	 (WORD_cast) self->parent_junct);
  fflush(stdout);
  
  while(curEle) {
    switch(curEle->child_type) {
    case CHILD_JUNCT:
      print_sem_constr_tree((sem_junct*)curEle->child, indent+1);
      fflush(stdout);
      break;
    case CHILD_ELEM:
      print_sem_elem_constr((sem_elem_constr*)curEle->child, indent+1);
      fflush(stdout);
      break;
    default:
      print_indentation(indent);
      printf("Wrong tag in sub_constr_list: %i", curEle->child_type);
      fflush(stdout);
    }
    curEle = curEle->next;
  }
}
   

/* Display all semantic constraint trees which have an elementary constraint
   with left side variable var */
void  print_sem_constr_trees_at_variable(var)
WORD*  var;
{
   sem_constr_list   root_list,root_list_ele;

   printf("print_sem_constr_trees_at_variable(");
   disp_(stdout,var,bp);
   printf("):\n");
   fflush(stdout);

   if (GETTAG(*var) != T_FVAR) {
      printf("Variable is already bound.\n");
      fflush(stdout);
      return;
   }

   root_list = compute_list_of_root_nodes_at_variable(var); 
   
   /* Print trees of computed root nodes and free root_list: */

   if (root_list == NULL) {
      printf("Variable has no living (alive) constraints.\n");
      fflush(stdout);
      return;
   }

   while (root_list != NULL) {
      root_list_ele = root_list;
      print_sem_constr_tree((sem_junct*) root_list_ele->child, 0);
      root_list = root_list->next;
      free(root_list_ele);
   }
}



/*************** procedures for the creation of the sem_constr-tree ******************/
/* sem-constr trees have to be built top-down, i.e. start with the root node
   and finish with the leaf nodes. Use add_new_sem_junct to create junctors
   (e.g. the root junctor) and use add_new_sem_elem_constr to create
   elementary constraints (i.e. the leaf nodes) */
void add_to_sub_constr_list(outParentJunct, child, type)
/* allocate new sem_constr_list element on the constraint stack,
 * initialize it (with child and type),
 * and insert it as first element into the sub_constr_list of
 * outParentJunct */ 
sem_junct*      outParentJunct; /* parent junctor */
sem_child*      child;
int             type;           /* type of the child: junctor or elemConstr */
{

  sem_constr_list new_ele = (sem_constr_list) c_sp;
  c_sp += SIZE_IN_WORDS(sem_constr_list_ele);
  
  /* check for stack-overflow */
  if (c_sp >= c_stack_end )
    sam_error("Constraint-Stack-Overflow", NULL, 2);

  /* initialize new_ele */
  new_ele->child_type = type;
  new_ele->child      = child;
  new_ele->next       = outParentJunct->sub_constr_list;

  /* trail the beginning of sub_constr_list and the arity before inserting new_ele */ 
  trail_ptr((void**) &(outParentJunct->sub_constr_list));
  trail_int(&(outParentJunct->arity));

  /* insert new_ele */
  outParentJunct->sub_constr_list = new_ele;
  outParentJunct->arity++;
  /* SEM CONSTRS TIGHTENED */
  if (!checking_global_consistency) {
     trail_ptr((void**) &(outParentJunct->chp_last_change));
     outParentJunct->chp_last_change = chp;
  }
}



sem_junct*  add_new_sem_junct(parentJunct, junctType)
/* create new sem_junct on the constraint-stack,
 * initialize it and add it to the sub_constr_list of parentJunct.
 * If parentJunct == NULL a new root node is being created */
sem_junct*     parentJunct;
int            junctType;
{
  sem_junct* outJunct;
  
  outJunct = (sem_junct *) c_sp;
  c_sp += SIZE_IN_WORDS(sem_junct);

  /* check for stack-overflow */
  if (c_sp >= c_stack_end )
    sam_error("Constraint-Stack-Overflow", NULL, 2);
  
  outJunct->parent_junct    = parentJunct;
  outJunct->sub_constr_list = NULL;
  outJunct->type            = junctType;
  outJunct->arity           = 0;
  outJunct->chp_last_change = chp;   /* SEM CONSTRS TIGHTENED */

  if (parentJunct)
    add_to_sub_constr_list(parentJunct, (sem_child*)outJunct, CHILD_JUNCT);

  return outJunct;
}



void add_new_sem_elem_constr(type, rightSide, parentJunct, outVar)
/* create new sem_elem_constr on the constraint-stack and initialize it
   insert it as first element to outVar->sem_equ_constr and add it to the
   sub_constr_list of parentJunct. */
int               type;
WORD*             rightSide;
sem_junct*        parentJunct;
WORD*             outVar;              
{
  sem_elem_constr* newElemConstr;

  newElemConstr = (sem_elem_constr *) c_sp;
  c_sp += SIZE_IN_WORDS(sem_elem_constr);
  
  /* check for stack-overflow */
  if (c_sp >= c_stack_end )
    sam_error("Constraint-Stack-Overflow", NULL, 2);

  newElemConstr->type = type;
  newElemConstr->right_side = rightSide;

  newElemConstr->parent_junct = parentJunct;
  newElemConstr->next         = ((constr_block *)GETPTR(*outVar))->sem_equ_constr;
  newElemConstr->var          = outVar;
  
  trail_ptr((void**) &(((constr_block*) GETPTR(*outVar))->sem_equ_constr));
  ((constr_block *)GETPTR(*outVar))->sem_equ_constr = newElemConstr;

  add_to_sub_constr_list(parentJunct, (sem_child*) newElemConstr, CHILD_ELEM);
}



/******* procedures for the attribute-handling in the sem_constr-tree ********/

sem_constr_list  compute_list_of_root_nodes_at_variable(var)
WORD*  var;
{
   sem_elem_constr*  eq_constr;
   sem_junct*        constr;
   sem_constr_list   root_list,root_list_ele,list_ptr;
   int               is_member;

   root_list = NULL;

   /* Get first elementary equality constraint to check: */
   eq_constr = ((constr_block*) GETPTR(*var))->sem_equ_constr;
   while (eq_constr != NULL) {
      if (!is_sem_elem_constr_alive(eq_constr))
	 eq_constr = eq_constr->next;
      else
	 break;
   }
   
   /* Process all elementary equality constraints of var: */
   while (eq_constr != NULL) {

      /* Get root node belonging to eq_constr: */
      constr = eq_constr->parent_junct;
      while (constr->parent_junct != NULL)
	 constr = constr->parent_junct;

      /* Check if found root node is already element of root_list: */
      is_member = FALSE;
      list_ptr = root_list;
      while (list_ptr != NULL  &&  !is_member) {
	 if ((sem_junct*) list_ptr->child == constr)
	    is_member = TRUE;
	 else
	    list_ptr = list_ptr->next;
      }

      /* If found root node is not yet element of root_list, insert it
	 in root_list: */
      if (!is_member) {
	 root_list_ele = (sem_constr_list) malloc(sizeof(sem_constr_list_ele));
	 root_list_ele->child = constr;
	 root_list_ele->next = root_list;
	 root_list = root_list_ele;
      }
      
      /* Get next elementary equality constraint to check: */
      eq_constr = eq_constr->next;
      while (eq_constr != NULL) {
	 if (!is_sem_elem_constr_alive(eq_constr))
	    eq_constr = eq_constr->next;
	 else
	    break;
      }
   }

   return root_list;
}



void remove_from_sub_constr_list(junct, child)
/* Remove child from junct's sub_constr_list 
 * child isn't lost because it has been trailed 
 */
sem_junct* junct;
sem_child* child;
{
  sem_constr_list predecessor = NULL;
  sem_constr_list curElement  = junct->sub_constr_list;

  if (junct->arity == 0)
    sam_error("remove_from_sub_constr_list: Arity is already 0",
	      NULL, 1);
  
  /* Find the predecessor of the element that has to be removed */
  while (curElement->child != child) {
    predecessor = curElement;
    curElement  = curElement->next;
    if (! curElement)
      sam_error("remove_from_sub_constr_list: child hasn't been found in sub_constr_list", NULL, 1);
  }
  
  /* is child the first element of the list? */
  if (! predecessor ) {
    trail_ptr((void**) &(junct->sub_constr_list));
    junct->sub_constr_list = curElement->next;
  }
  else {
    trail_ptr((void**) &(predecessor->next));
    predecessor->next = curElement->next;
  }
  
  trail_int(&(junct->arity));
  junct->arity--;
}



/* is_sem_junct_alive
 * return TRUE, if the constraint is neither deleted nor deactivated 
 * else      mark the whole branch of this constraint and return FALSE
 */ 
int is_sem_junct_alive(myJunct)
sem_junct*  myJunct;
{
  sem_junct*  endJunct;
  int         myJunctType;
  
  endJunct = myJunct;

  /* find the marked ancestor and return TRUE if there isn't one */ 
  while (! ((endJunct->type == DELETED) || (endJunct->type == DEACTIVATED)) ) {
    endJunct = endJunct->parent_junct;
    if (endJunct == NULL)
      return TRUE;
  }

  /* mark this branch as deleted or deactivated, respectively */
  myJunctType = endJunct->type;

  while (myJunct != endJunct) {
    trail_int(&(myJunct->type));
    trail_int(&(myJunct->arity));
    trail_ptr((void**) &(myJunct->sub_constr_list));
    myJunct->type            = myJunctType;
    myJunct->arity           = 0;
    myJunct->sub_constr_list = NULL;
    myJunct = myJunct->parent_junct;
  }

  return FALSE;
}



/* is_sem_elem_constr_alive
 * return TRUE, if the constraint is neither deleted nor deactivated 
 * else      mark the whole branch of this constraint and return FALSE
 */ 
int is_sem_elem_constr_alive(elemConstr)
sem_elem_constr*  elemConstr;
{
  sem_junct* myJunct = elemConstr->parent_junct; /* Cannot be NULL */
  
  if ((elemConstr->type == DEACTIVATED) || (elemConstr->type == DELETED))
    return FALSE;

  if (is_sem_junct_alive(myJunct))
    return TRUE;   /* elemConstr and its parentJuncts are alive */

  /* One of the parents is not alive */
  trail_int(&(elemConstr->type));  
  elemConstr->type = myJunct->type; /* Use value that has been propagated downward by is_sem_junct_alive */

  return FALSE;
}



/* Deactivate a living (alive) elemConstr, as it can't be violated any more
 * and propagate deactivation upward, i.e. deactivate disjunctions in which
 * one child is deactivated and deactivate conjunctions in which all children
 * are deactivated */
void deactivate_sem_elem_constr(elemConstr)
sem_elem_constr* elemConstr;
{
  sem_junct* curJunct;
  sem_child* curChild;

  /* deactivate the elementary constraint */
  trail_int(&(elemConstr->type));
  elemConstr->type = DEACTIVATED;

  /* deactivate its ancestors (if necessary) */
  curChild = (sem_child*) elemConstr;
  curJunct = elemConstr->parent_junct;

  while (curJunct) {
    
     switch (curJunct->type) {

     case JUNCT_DISJ:
	   trail_int(&(curJunct->arity));
	   trail_ptr((void**) &(curJunct->sub_constr_list));
	   curJunct->arity           = 0;
	   curJunct->sub_constr_list = NULL;
	   break;

     case JUNCT_CONJ:
	   /* Remove curChild from sub_constr_list of curJunct and
	      decrement arity of curJunct by 1: */
	   remove_from_sub_constr_list(curJunct, curChild);
	   if ( curJunct->arity > 0 ) 
	      /* in this case deactivation doesn't have to be
		 propagated upward */ 
	      return;
	   break;

     default:
	   sam_error("deactivate_sem_elem_constr: Wrong junctor type.",
		     NULL, 3);

     } /* switch */
    
     trail_int(&(curJunct->type));
     curJunct->type = DEACTIVATED;
     
     /* go upward */
     curChild = (sem_child*) curJunct;
     curJunct = curJunct->parent_junct;
     
  } /* while */
}



/* Delete a living (alive) elemConstr, as it can't be fulfilled any more
 * and propagate deletion upward, i.e. delete conjunctions in which
 * one child is deleted and delete disjunctions in which all children
 * are deleted 
 *
 * returns FALSE, if root node has been deleted, i.e. the constraint net
              is inconsistent 
 * returns TRUE  otherwise
 */
int  delete_sem_elem_constr(elemConstr)
sem_elem_constr*  elemConstr;
{
  sem_junct* curJunct;
  sem_child* curChild;

  /* delete the elementary constraint */
  trail_int(&(elemConstr->type));
  elemConstr->type = DELETED;
  
  /* delete its ancestors (if necessary) */
  curChild = (sem_child*) elemConstr;
  curJunct = elemConstr->parent_junct;

  while (curJunct) {

     /* SEM CONSTRS TIGHTENED */
     if (!checking_global_consistency) {
	trail_ptr((void**) &(curJunct->chp_last_change));
	curJunct->chp_last_change = chp;
     }

     switch (curJunct->type) {

     case JUNCT_DISJ:
	   /* Remove curChild from sub_constr_list of curJunct and
	      decrement arity of curJunct by 1: */
	   remove_from_sub_constr_list(curJunct, curChild);
	   if ( curJunct->arity > 1 ) {
	      /* in this case deletion doesn't have to be propagated
		 upward */
	      return TRUE;
	   }
	   if ( curJunct->arity == 1 ) {
	      /* in this case try deterministic instantiation */
	      return try_det_inst(curJunct);
	   }
	   break;

     case JUNCT_CONJ:
	   trail_int(&(curJunct->arity));
	   trail_ptr((void**) &(curJunct->sub_constr_list));
	   curJunct->arity           = 0;
	   curJunct->sub_constr_list = NULL;
	   break;

     default:
	   sam_error("delete_sem_elem_constr: Wrong junctor type.",
		     NULL, 3);

     } /* switch */
	   
     trail_int(&(curJunct->type));
     curJunct->type = DELETED;
     
     /* go upward */
     curChild = (sem_child*) curJunct;
     curJunct = curJunct->parent_junct;
     
  } /* while */
  
  return FALSE;
}



/* generate new elementary subconstraints from bindings, which 
 * have been written on the PU-trail during the protocol unification.
 */
void generate_sem_elem_constraints_from_pu_trail(parentJunct, type)
sem_junct*        parentJunct;    /* parent junctor of new elem constraints */
int               type;         
{
  WORD*             left_gp;        /* left side variable of new generated elem. constr. */
  WORD*             right_gp;       /* right side term of new generated elem. constr. */
  s_trail*          tpt;

  /***********************************************************/
  /* look in the PU-trail, which variables were bound by the */
  /* protocol-unification                                    */
  /***********************************************************/
  for (tpt = pu_tr-1; tpt >= pu_trail; tpt--) {
    
    /************************************************************/
    /* left_gp points to the T_FVAR that was bound by protocol  */
    /* unification. Look for what it's been bound to and create */
    /* a constraint                                             */
    /************************************************************/
    left_gp  = (WORD *)tpt->ptr;
    right_gp = deref(left_gp, NULL);
    
    *(tpt->ptr) = tpt->oval;           /* untrail the bound var */

    add_new_sem_elem_constr(type, right_gp, parentJunct, left_gp);
  }
       
  pu_tr = pu_trail;                    /* reset pu_trail */
}



/* decompose elementary constraints which still can be fulfilled or violated
 * i.e. generate new elementary subconstraints from bindings, which 
 * have been written on the PU-trail during the protocol unification.
 * The original elementary constraint will be deactivated 
 */
void decompose_sem_elem_constr(elemConstr)
sem_elem_constr*  elemConstr;     /* elem. constr. to be decomposed */
{
   generate_sem_elem_constraints_from_pu_trail(elemConstr->parent_junct,
					       elemConstr->type); 
   deactivate_sem_elem_constr(elemConstr);
}



/*------------------------- deterministic instantiation -----------------*/

/* Return TRUE if no disjunction with more than one subconstraint is
            found on the way up through the tree
   Return FALSE otherwise
*/
int is_sem_junct_upwards_deterministic(myJunct)
sem_junct* myJunct;
{
   while (myJunct) {
      if ((myJunct->type == JUNCT_DISJ)  &&  (myJunct->arity > 1))
	 return FALSE;
      myJunct = myJunct->parent_junct;
   }

   return TRUE;
}



/* inserts elemConstr as first element in InputList and returns the
   changed list. 
*/
sem_constr_list   insert_elem_constr(elemConstr,InputList)
sem_elem_constr*  elemConstr;
sem_constr_list   InputList;
{
   sem_constr_list   List;
   
   List = (sem_constr_list) malloc(sizeof(sem_constr_list_ele));
   
   List->child_type = CHILD_ELEM;
   List->child      = elemConstr;
   List->next       = InputList;
   
   return List;
}



/* Returns a list of all deterministic, living (alive) elementary
   constraints which are leafs in the tree with the root Junct.
   */
sem_constr_list   get_elem_constr_list(Junct,InputDetInstList)
sem_junct*        Junct;
sem_constr_list   InputDetInstList;
{
   sem_constr_list   DetInstList, SubConstr;
   int               ChildType;
   sem_junct*        ChildJunct;
   sem_elem_constr*  ChildElem;

   switch (Junct->type) {
      
   case JUNCT_DISJ:

	 if (Junct->arity > 1)
	    /* Junct is not downwards deterministic */
	    return InputDetInstList;
	 
	 /* Junct->arity == 1
	    i.e., Junct is downwards deterministic */

	 ChildType = Junct->sub_constr_list->child_type;

	 if (ChildType == CHILD_JUNCT) {
	    ChildJunct = (sem_junct*) Junct->sub_constr_list->child;
	    if (ChildJunct->type != DEACTIVATED  &&
		ChildJunct->type != DELETED) 
	       return get_elem_constr_list(ChildJunct, InputDetInstList);
	 }
	 else { /* ChildType == CHILD_ELEM */
	    ChildElem = (sem_elem_constr*) Junct->sub_constr_list->child;
	    if (ChildElem->type != DEACTIVATED  &&
		ChildElem->type != DELETED)
	       return insert_elem_constr(ChildElem, InputDetInstList);
	 }
	 
	 /* Child is DEACTIVATED or DELETED */
	 return InputDetInstList;
	    
   case JUNCT_CONJ:
	 
	 DetInstList = InputDetInstList;
	 SubConstr = Junct->sub_constr_list;
	 
	 while (SubConstr != NULL) {

	    ChildType = SubConstr->child_type;

	    if (ChildType == CHILD_JUNCT) {
	       ChildJunct = (sem_junct*) SubConstr->child;
	       if (ChildJunct->type != DEACTIVATED  &&
		   ChildJunct->type != DELETED)
		  DetInstList = get_elem_constr_list(ChildJunct,
						     DetInstList); 
	    }
	    else { /* ChildType == CHILD_ELEM */
	       ChildElem = (sem_elem_constr*) SubConstr->child;
	       if (ChildElem->type != DEACTIVATED  &&
		   ChildElem->type != DELETED)
		  DetInstList = insert_elem_constr(ChildElem,
						   DetInstList); 
	    }

	    SubConstr = SubConstr->next;
	 }
	 
	 return DetInstList;
	 
   default: sam_error("get_elem_constr_list: Wrong junctor type.",
		      NULL, 2);
	 
   }

   /* gcc doesn't recognize that every case has a return. The
      following return has been added to avoid a gcc warning: */
   return InputDetInstList;
}



void  free_det_inst_list(list)
sem_constr_list  list ;
{
   sem_constr_list  next;

   while (list != NULL) {
      next = list->next;
      free((void*) list);
      list = next;
   }
}



/* Tries deterministic instantiation for the elementary constraint
   myElem:
   returns FALSE if deterministic instantiation failed
   returns TRUE otherwise, i.e., if deterministic instantiation was
             successful or if deterministic instantiation could not be
	     tried because myElem was not alive
 */
#define FAIL_U       {return FALSE;}
#define UNIF_FAIL(x) FAIL_U

int try_det_inst_elem(myElem)
sem_elem_constr* myElem;
{
   WORD*            derefRightSide;
   
   if (is_sem_elem_constr_alive(myElem)) {
      
      switch (myElem->type) {
	 
      case ELEM_EQU:

	    /* try to make the deterministic instantiation 
	       UNIFY calls macros like UN_FVAR, and UN_FVAR calls
	       FAIL_U or UNIF_FAIL if not successful */ 
	    derefRightSide = deref(myElem->right_side, bp);
	    UNIFY(myElem->var, bp, derefRightSide, bp);

	    /* the instantiation was successful */
	    return TRUE;
	    
      case ELEM_DIFF:

	    sam_error("try_det_inst_elem: disequality constraints are not yet implemented.", 
		      NULL, 2);

	    pu_tr = pu_trail;
	    if (un_protoco(myElem->var, bp, myElem->right_side, bp, 1)) {
	       /* any unifications made? */
	       if (pu_tr == pu_trail) {
		  /* var and right_side are equal
		   * delete myElem, for it cannot be fulfilled any more 
		   * if the root-node is deleted, a fail has to follow
		   */
		  return delete_sem_elem_constr(myElem);
	       }
	       else {
		  /* decompose the constraint net with the
		     unifications on the pu_trail */ 
		  decompose_sem_elem_constr(myElem);
		  return TRUE;
	       }
	    } /* if unprotoco */
	    else {
	       /* no unifications are possible, therefore the
		  constraint can't be violated any more */
	       pu_untrail(); 
	       deactivate_sem_elem_constr(myElem);
	       return TRUE;
	    }
      } /* switch */
   }
   return TRUE;
}



/* Tries deterministic instantiation at a junctor Junct:
   returns FALSE if deterministic instantiation was not successful
   returns TRUE otherwise, i.e., if deterministic instantiation was
             successful or if deterministic instantiation has not been
	     tried because Junct was not upwards deterministic
 */
int  try_det_inst(Junct)
sem_junct*  Junct;
{
   sem_elem_constr*  ElemConstr;
   sem_constr_list   ElemConstrListPtr, ElemConstrList;
      
   if (is_sem_junct_upwards_deterministic(Junct->parent_junct)) { 
      
      ElemConstrList = get_elem_constr_list(Junct, NULL); 
      ElemConstrListPtr = ElemConstrList;
      
      while (ElemConstrListPtr != NULL) {
	 ElemConstr = (sem_elem_constr*) ElemConstrListPtr->child;
	 if (try_det_inst_elem(ElemConstr) == 0) { 
	    free_det_inst_list(ElemConstrList);
	    return FALSE;
	 }
	 ElemConstrListPtr = ElemConstrListPtr->next;
      }
      
      free_det_inst_list(ElemConstrList);
      
   } /* if (is_sem_junct_upwards_deterministic(...) */
   
   return TRUE;
}
