% -*- prolog -*-

% -----------------------------------------------------------
% equconstr/4
% generates constraints of the form "$equconstr( or(and([[V1, V2, ...], [a, b, ...]]), and([[V1, ...], [c, ...]]) ... ), N)" 
% which are processed by the built-in defined in sam/i_equconstr.c
% equconstr/4 always succeeds
% ----------------------------------------------------------- 

equconstr(problemwise, [], _, []).

equconstr(problemwise, InputClauseList, _, OutputClauseList) :-

    copy_clauses(InputClauseList, NewInputClauseList),
    remove_subsumed_clauses(NewInputClauseList, RestOfInputClauseList),
    setval(rename_counter, 1),
    gen_equconstrs(RestOfInputClauseList, OutputClauseList).


% -----------------------------------------------------------
% process_switch/1
% -----------------------------------------------------------

process_switch_equconstr( ["-equconstr" | ArgList] ) :-
	
	(occurs(onlyfacts, ArgList) ->
	    Argument1 = onlyfacts
	;
	    Argument1 = []
	),

	((occurs(nonNF, ArgList) ; occurs(opt, ArgList)) ->
	    Argument2 = nonNF,
	    (occurs(decomposeVLC, ArgList) ->
		Argument3 = decomposeVLC
	    ;
	        (occurs(decomposeLVC, ArgList) ->
		    Argument3 = decomposeLVC
		;
		    (occurs(decomposeLCV, ArgList) ->
			Argument3 = decomposeLCV
		    ;
		        Argument3 = []
		    )
		)
	    )
	;
	    Argument2 = [],
	    Argument3 = []
	),
	
	(occurs(noperm, ArgList) ->
	    Argument4 = noperm
	;
	    Argument4 = []
	),
	
	assert( cmd_arg(equconstr, (Argument1, 
	                            Argument2, 
				    Argument3, 
				    Argument4)) ),
	write("# *** option: equconstr ("),
	write(Argument1),
	write(", "),
	write(Argument2),
	write(", "),
	write(Argument3),
	write(", "),
	write(Argument4),
	writeln(")"),
	flush(output).


% -----------------------------------------------------------
% gen_disconstrs([Clause | InputClauseList], 
%                [NewClause|LocalOutputClauseList])
% gen_equconstrs/2 finds all clauses in InputClauseList, 
% which together with Clause can be summarized as the single 
% clause NewClause.
% NewClause is added to the output clause list (2nd arg).
% This is done recursively for all clauses in InputClauseList.
% ------------------------------------------------------------  

gen_equconstrs([], []). 


gen_equconstrs([Clause | InputClauseList], [NewClause|LocalOutputClauseList]) :-

    % find all the clauses suitable for the summarization 
    % and put them into ListOfSummarizableClauses.
    % RestOfInputClauseList is the rest of InputClauseList,
    % i.e. InputClauseList \ ListOfSummarizableClauses
    get_summarizable_entities_equconstr(Clause, 
			                InputClauseList,
					ListOfSummarizableClauses,
					RestOfInputClauseList),

    % complete the summarization into NewClause
    summarize_clauses_equconstr([Clause | ListOfSummarizableClauses],
		                NewClause),
    
    % proceed recursively
    gen_equconstrs(RestOfInputClauseList, 
		   LocalOutputClauseList).



% ------------------------------------------------------------------
% get_summarizable_entities_equconstr(Entity, 
%                                     InputList, 
%                                     SummarizableEntities, 
%                                     RestOfInputList) 
% finds all the entities (facts and/or clauses) in InputList which 
% can be summarized with Entity and writes them into SummarizableEntities.
% RestOfInputList = InputList \ SummarizableEntities
%
% Two entities are summarizable, if 
% 1. they've got the same number of literals
% 2. they've got the same set of predicates (same_predicates/3)
% -------------------------------------------------------------------

% The entity is a clause. Only try to summarize if the onlyfacts option
% cmd_arg(equconstr, (onlyfacts, _, _, _)) is NOT used
get_summarizable_entities_equconstr(InputClause, 
                                    InputList, 
				    SummarizableEntities, 
				    RestOfInputList) :-

	cmd_arg(equconstr, ([], _, _, _)),    % NOT option onlyfacts
	get_summarizable_clauses(InputClause, 
                                 InputList, 
				 SummarizableEntities, 
				 RestOfInputList).


get_summarizable_entities_equconstr(InputClause, 
                                    InputList, 
				    SummarizableEntities, 
				    RestOfInputList) :-

	cmd_arg(equconstr, (onlyfacts, _, _, _)),
	InputClause =.. [_, _, _, LiteralList, _, _],
	fact(LiteralList),
	get_summarizable_clauses(InputClause, 
	                         InputList, 
				 SummarizableEntities, 
				 RestOfInputList).


% Otherwise "return" empty SummarizableEntities (entity is either 
% a rule or not a fact & the onlyfacts-mode is selected)
get_summarizable_entities_equconstr(_, InputList, [], InputList).

	      

% ------------------------------------------------------------------
% summarize_clauses_equconstr(SummarizableClauses, OutputClause)
% Summarize SummarizableClauses into OutputClause, 
% which may contain a Literal (actually a constraint) "--$equconstr(...)"
% ------------------------------------------------------------------

summarize_clauses_equconstr([OutputClause],OutputClause).

summarize_clauses_equconstr(SummarizableClauses, OutputClause) :-

    % get list of arglists and append dictionaries
    % of the clauses in SummarizableClauses
    get_list_of_arglists(SummarizableClauses, ListOfArgLists),
    append_dictionaries(SummarizableClauses, AppendedDictionaries),

    % prefix the output names of all variables in order to avoid 
    % name-clashes with new generated variables Vi or Xi
    rename_variables(AppendedDictionaries, RenamedAppendedDictionaries),

    % with the help of this list, generate OutputClause from HeadClause
    SummarizableClauses = [HeadClause |_],

    gen_new_clause_equconstr(HeadClause,
	                     ListOfArgLists,
			     RenamedAppendedDictionaries, 
			     OutputClause).
	


% ------------------------------------------------------------------
% gen_new_clause_equconstr(ExampleClause, ListOfArgLists, NewClause)
% Generates NewClause from ExampleClause and ListOfArgLists by renaming 
% all of the arguments and possibly adding a constraint:
% NewClause will contain a constraint ($equconstr), if ListOfArgLists 
% isn't empty.
% ------------------------------------------------------------------

% ListOfArgLists is empty
gen_new_clause_equconstr(ExampleClause, [], _, ExampleClause).


% ListOfArgList isn't empty
gen_new_clause_equconstr(InClause, 
	                 ListOfArgLists, 
			 AppendedDictionaries, 
			 OutClause) :- 

    InClause =.. [U, V, W, LiteralList, X, _],

    % rename all of the n arguments in LiteralList to V1, ..., Vn 
    % and return NewLiteralList and VarList = [V1, ..., Vn] 
    gen_new_literallist(LiteralList, ListOfArgLists, 
	                1, [],
                        NewLiteralList, NewListOfArgLists, 
                        VarList),

    % generate the constraint from ListOfArgLists and VarList
    gen_equ_constraint(NewListOfArgLists, 
                       VarList, 
                       EqConstraint),

    check_dictionary([EqConstraint, NewLiteralList], AppendedDictionaries),

    % finally generate OutClause
    length(ListOfArgLists, NrOfSummarizedClauses),
    OutClause =.. [U, V, W, 
	           [--'$equconstr'(EqConstraint, NrOfSummarizedClauses) | NewLiteralList],
	           X, AppendedDictionaries].
    


% true if all variables of Term occur in Dictionary
check_dictionary(Term, Dictionary) :-
	term_variables(Term, VarList),
	check_dictionary_help(VarList, Dictionary).

check_dictionary_help([], _).

check_dictionary_help(VarList, Term) :-
	VarList = [Head | Tail],
	(occurs(Head, Term) ->
	    check_dictionary_help(Tail, Term)
	;
	    !,
	    write('ERROR: variable '),
	    write(Head),
	    writeln(' does not occur in dictionary'),
	    flush(output),
	    fail
	).
	



% ------------------------------------------------------------------
% gen_equ_constraint(ListOfArgLists, VarList, Constraint)
% generate Constraint from ListOfArglists and VarList
% ListOfArgLists = [ArgList_1, ArgList_2, ..., ArgList_n]
% Constraint     = sem_or(sem_and([VarList, ArgList_1]), 
%                         sem_and([VarList, ArgList_2]), 
%                         ..., 
%                         sem_and([VarList, ArgList_n]))
% ------------------------------------------------------------------

gen_equ_constraint(ListOfArgLists, VarList, OutConstraint) :-

	gen_subconstraints(ListOfArgLists, VarList, SubConstraints),
	length(SubConstraints, L),
	(L == 1 -> 
	    OutConstraint =.. SubConstraints
	; 
	    (cmd_arg(equconstr, (_, nonNF, _, _)) -> 
		simplify(SubConstraints, OutConstraint, _)
	    ;
		OutConstraint =.. [sem_or | SubConstraints]
	    )
	).
		

gen_subconstraints([], _, []).

gen_subconstraints([ArgList | TailOfArgList], 
                   VarList, 
		   [sem_and([VarList, ArgList]) | TailOfConstraintList]) :-
   gen_subconstraints(TailOfArgList, VarList, TailOfConstraintList).



% ---------------------- Simplification -------------------
% ---------------------------------------------------------


% -----------------------------------------------------------------
% simplify(ListOfANDConstraints, OutConstraint, Depth)
%
% ListOfANDConstraints is a (disjunctive) list of conjunctions. 
% Try to simplify ListOfANDConstraints by reducing duplicates, e.g.:
% [( A = a    &  B = b1  &  C = c1 ),
%  ( A = a    &  B = b2  &  C = c2 ),
%  ( A = a    &  B = b1  &  C = c3 ),
%  ( A = aaa  &  B = b1  &  C = c1 )]
% can be simplified to 
% (  (A = a & (  (B = b1 & C = c1)
%              v (B = b2 & C = c2) 
%              v (B = b1 & C = c3))) 
%  v (A = aaa & B = b1 & C = c1))
% which can be simplified furthermore to
% (  (A = a & (  (B = b1 & (  C = c1 
%                           v C = c3))
%              v (B = b2 & C = c2))) 
%  v (A = aaa & B = b1 & C = c1))
%
% Use Dictionary for comparisons during constraint simplification.
% -----------------------------------------------------------------

simplify([], _, _) :-
	!,
	writeln('ERROR in simplify: empty list of AND constraints'),
	flush(output),
	fail.


simplify(ListOfANDConstraints, OutConstraint, ORDepth) :-

	get_simplification(ListOfANDConstraints,
	                   Position,
			   SimplifiableConstraints,
			   Generalization,
			   TypeOfSimplification,
			   RestOfListOfANDConstraints),

	(Position == 0 ->

	    get_cascaded_disjunction(RestOfListOfANDConstraints,
	                             OutConstraint),
	    ORDepth = 1
	;
	    apply_simplification(Position, 
	                         SimplifiableConstraints,
				 Generalization,
				 TypeOfSimplification,
	                         ElemConstr,
				 ChangedConstraints),
	    simplify(ChangedConstraints, 
		     ComplConstr, 
		     D1),
	    SimplifiedConstraint =.. [sem_and, ElemConstr, ComplConstr],
	    
	    (RestOfListOfANDConstraints == [] ->
		
		OutConstraint = SimplifiedConstraint,
		ORDepth = D1
	    ;
		simplify(RestOfListOfANDConstraints, 
		         SimplifiedRestConstr,
			 D2),
		compose(sem_or, 
		        SimplifiedConstraint,
			SimplifiedRestConstr,
			D1,
			D2,
			OutConstraint),
		max(D1, D2, D3),
		ORDepth is D3 + 1
	    )
	).



get_simplification(ConstraintList,
		   Position,
	           SimplifiableConstraints,
		   Generalization,
		   TypeOfSimplification,
		   RestOfConstraintList) :-

	ConstraintList = [sem_and([VarList, _]) | _],
	length(VarList, L),
	get_best_simplification(ConstraintList,
	                        L, 
				Simplification,
				TypeOfSimplification),
	Simplification = [Position, 
	                  SimplifiableConstraints,
			  Generalization,
			  RestOfConstraintList].

	    

get_best_simplification(ConstraintList, 
			0, 
			[0, [], _, ConstraintList],
			none).

get_best_simplification(ConstraintList, 
			VarCounter, 
			BestSimplification, 
			TypeOfBestSimplification) :-

	look_for_repeated_rows(VarCounter, 
	                       ConstraintList, 
			       LocalSimplification1),
	look_for_equal_top_symbols(VarCounter, 
	                           ConstraintList, 
				   LocalSimplification2),

	NewVarCounter is VarCounter - 1,
	get_best_simplification(ConstraintList, 
	                        NewVarCounter, 
				LocalSimplification3,
				TypeOfLocalSimplification3),

	(is_better_simplification_than(LocalSimplification1,
	                               LocalSimplification2) ->
		LocalBestSimplification       = LocalSimplification1,
		TypeOfLocalBestSimplification = repeatedrows
	    ;
	        LocalBestSimplification       = LocalSimplification2,
		TypeOfLocalBestSimplification = equaltopsymbols
	),
	(is_better_simplification_than(LocalBestSimplification,
	                               LocalSimplification3) ->
		BestSimplification       = LocalBestSimplification,
		TypeOfBestSimplification = TypeOfLocalBestSimplification
	    ;
		BestSimplification       = LocalSimplification3,
		TypeOfBestSimplification = TypeOfLocalSimplification3
	).



is_better_simplification_than(Simplification1, Simplification2) :-
	cmd_arg(equconstr, (_, _, decomposeVLC, _)),
	!,
	Simplification1 = [_, SimplConstrs1, Generalization1, _],
	Simplification2 = [_, SimplConstrs2, Generalization2, _],
	term_variables(Generalization1, VarList1),
	term_variables(Generalization2, VarList2),
	length(VarList1, V1),
	length(VarList2, V2),
	length(SimplConstrs1, L1),
	length(SimplConstrs2, L2),
	number_of_symbols_in_term(Generalization1, C1),
	number_of_symbols_in_term(Generalization2, C2),
	!,
	L1 > 0,         % if L1==0, then Simplification1 is not a simplification
	(
	    V1 < V2     % less new variables
	;
	    V1 == V2,
	    L1 > L2     % more simplifiable constraints
	;
	    V1 == V2,
	    L1 == L2,
	    C1 > C2     % higher complexity
	).

is_better_simplification_than(Simplification1, Simplification2) :-
	cmd_arg(equconstr, (_, _, decomposeLVC, _)),
	!,
	Simplification1 = [_, SimplConstrs1, Generalization1, _],
	Simplification2 = [_, SimplConstrs2, Generalization2, _],
	term_variables(Generalization1, VarList1),
	term_variables(Generalization2, VarList2),
	length(VarList1, V1),
	length(VarList2, V2),
	length(SimplConstrs1, L1),
	length(SimplConstrs2, L2),
	number_of_symbols_in_term(Generalization1, C1),
	number_of_symbols_in_term(Generalization2, C2),
	!,
	L1 > 0,         % if L1==0, then Simplification1 is not a simplification
	(
	    L1 > L2     % more simplifiable constraints
	;
	    L1 == L2,
	    V1 < V2     % less new variables
	;
	    L1 == L2,
	    V1 == V2,
	    C1 > C2     % higher complexity
	).

is_better_simplification_than(Simplification1, Simplification2) :-
	cmd_arg(equconstr, (_, _, decomposeLCV, _)),
	!,
	Simplification1 = [_, SimplConstrs1, Generalization1, _],
	Simplification2 = [_, SimplConstrs2, Generalization2, _],
	term_variables(Generalization1, VarList1),
	term_variables(Generalization2, VarList2),
	length(VarList1, V1),
	length(VarList2, V2),
	length(SimplConstrs1, L1),
	length(SimplConstrs2, L2),
	number_of_symbols_in_term(Generalization1, C1),
	number_of_symbols_in_term(Generalization2, C2),
	!,
	L1 > 0,         % if L1==0, then Simplification1 is not a simplification
	(
	    L1 > L2     % more simplifiable constraints
	;
	    L1 == L2,
	    C1 > C2     % higher complexity
	;
	    L1 == L2,
	    C1 == C2,
	    V1 < V2     % less new variables
	).

% option decompose is not switched on:
is_better_simplification_than(Simplification1, Simplification2) :-
	Simplification1 = [_, SimplConstrs1, Generalization1, _],
	Simplification2 = [_, SimplConstrs2, Generalization2, _],
	length(SimplConstrs1, L1),
	length(SimplConstrs2, L2),
	number_of_symbols_in_term(Generalization1, C1),
	number_of_symbols_in_term(Generalization2, C2),
	!,
	L1 > 0,         % if L1==0, then Simplification1 is not a simplification
	(
	    L1 > L2     % more simplifiable constraints
	;
	    L1 == L2,
	    C1 > C2     % higher complexity
	).


number_of_symbols_in_term(Term, 0) :-
	var(Term).
	 
number_of_symbols_in_term(Term, 0) :-
	atom(Term),
	(name(Term, [86 | _]) ; name(Term, [88 | _])).

number_of_symbols_in_term(Term, Number) :-
	\+ var(Term),
	Term =.. [_ | Args],
	number_of_symbols_in_termlist(Args, LocalNumber),
	Number is LocalNumber + 1.


number_of_symbols_in_termlist([], 0).

number_of_symbols_in_termlist(TermList, Number) :-
	TermList = [Head | Tail],
	number_of_symbols_in_term(Head, LocalNumber1),
	number_of_symbols_in_termlist(Tail, LocalNumber2),
	Number is LocalNumber1 + LocalNumber2.
	


compute_generalization(0, _, _) :-
	!,
	writeln("ERROR: cannot compute generalization"),
	flush(output),
	fail.

compute_generalization(_, [], _) :-
	!,
	writeln("ERROR: cannot compute generalization"),
	flush(output),
	fail.

compute_generalization(VarIndex, SimplConstrs, Generalization) :-
	!,
	copy_term(SimplConstrs, CopiedSimplConstrs),
	get_list_of_generalizable_terms1(VarIndex, 
	                                 CopiedSimplConstrs, 
					 ListOfGeneralizableTerms),
	generalize_termlists(ListOfGeneralizableTerms, 
	                     Generalization).



get_list_of_generalizable_terms1(_, [], []).

get_list_of_generalizable_terms1(VarIndex, 
	                         SimplConstrs, 
				 ListOfGeneralizableTerms) :-
	SimplConstrs = [sem_and([_, TermList]) | Tail],
	get_term(VarIndex, TermList, Term),
	get_list_of_generalizable_terms1(VarIndex, 
	                                 Tail, 
					 LocalListOfGeneralizableTerms),
	ListOfGeneralizableTerms = [Term | LocalListOfGeneralizableTerms].


get_list_of_generalizable_terms2(_, [], []).

get_list_of_generalizable_terms2(VarIndex, 
	                         ListOfArgLists, 
				 ListOfGeneralizableTerms) :-

	ListOfArgLists = [Head | Tail],
	get_term(VarIndex, Head, Term),
	get_list_of_generalizable_terms2(VarIndex, 
	                                 Tail, 
					 LocalListOfGeneralizableTerms),
	ListOfGeneralizableTerms = [Term | LocalListOfGeneralizableTerms].



generalize_termlists([], _) :-
	!,
	writeln("ERROR: cannot compute generalization"),
	flush(output),
	fail.

generalize_termlists(ListOfGeneralizableTerms, Generalization) :-
	(get_common_top_symbol(ListOfGeneralizableTerms,
	                       CommonTopSymbol,
			       _,
			       ListOfArgLists) ->
		(var(CommonTopSymbol) ->
		    !,
		    writeln('ERROR: CommonTopSymbol is a variable'),
		    flush(output),
		    fail
		;   
		    generalize_arglists(ListOfArgLists, 
		                        GeneralizedArgs),
		    Generalization =.. [CommonTopSymbol | GeneralizedArgs]
		)
	;
	        getnewvar(Generalization)
	).



% get_common_top_symbol(ListOfGeneralizableTerms,
% 	                CommonTopSymbol,
%                       RenameFlag,
% 			ListOfArgLists)
% CommonTopSymbol is common top symbol of all elements in 
% ListOfGeneralizableTerms.
% ListOfArgLists is a list which contains the arglists of all elements in 
% ListOfGeneralizableTerms.
%
% - if all elements in ListOfGeneralizableTerms are variables, 
%   then CommonTopSymbol is a renamed variable (RenameFlag indicates renaming)
% - if all elements in ListOfGeneralizableTerms begin with the same function 
%   symbol, then CommonTopSymbol is the respective function symbol
% - otherwise, get_common_top_symbol fails
%
% clause 1: ListOfGeneralizableTerms must not be empty
% clause 2: handels the case where ListOfGeneralizableTerms has only 1 element
% clause 3: checks the head of ListOfGeneralizableTerms and makes a 
%           recursive call
%
get_common_top_symbol([], _, _, []) :-
	!,
	writeln("ERROR: cannot get common top symbol"),
	flush(output),
	fail.

get_common_top_symbol([Term], CommonTopSymbol, RenameFlag, [ArgList]) :-
	!,
	(var(Term) ->
	    rename_varlist([_],[CommonTopSymbol]),
	    RenameFlag = true,
	    Term = CommonTopSymbol,
	    ArgList = []
	;
	    Term =.. [CommonTopSymbol | ArgList],
	    RenameFlag = false
	).

get_common_top_symbol(ListOfGeneralizableTerms,
	              CommonTopSymbol,
		      RenameFlag,
		      ListOfArgLists) :-
	ListOfGeneralizableTerms = [Head | Tail],
	get_common_top_symbol(Tail,
	                      CommonTopSymbol,
			      RenameFlag,
			      LocalListOfArgLists),
	(var(Head) ->
	    RenameFlag == true,
	    Head = CommonTopSymbol,
	    ListOfArgLists = [[] | LocalListOfArgLists]
	;
	    \+ var(CommonTopSymbol),
	    Head =.. [CommonTopSymbol | ArgList],
	    ListOfArgLists = [ArgList | LocalListOfArgLists]
	).
	    


generalize_arglists([], []) :-
	!,
	writeln("ERROR: cannot generalize arglists"),
	flush(output),
	fail.

generalize_arglists(ListOfArgLists, GeneralizedArgs) :-
	ListOfArgLists = [Head | _],
	length(Head, Length),
	(Length == 0 ->
	    GeneralizedArgs = []
	;
	    generalize_arglists_help(Length, ListOfArgLists, GeneralizedArgs)
	).



generalize_arglists_help(_, [], [])  :-
	!,
	writeln("ERROR: does never occur"),
	flush(output),
	fail.

generalize_arglists_help(0, _, []) :-
	!,
	writeln("ERROR: trying to generalize 0-th elements of arglists"),
	flush(output),
	fail.

generalize_arglists_help(1, ListOfArgLists, [Generalization]) :-
	get_list_of_generalizable_terms2(1, 
	                                 ListOfArgLists, 
					 ListOfGeneralizableTerms),
	generalize_termlists(ListOfGeneralizableTerms, 
	                     Generalization).

generalize_arglists_help(VarIndex, ListOfArgLists, GeneralizedArgs) :-
	NewVarIndex is VarIndex - 1,
	generalize_arglists_help(NewVarIndex, 
	                         ListOfArgLists, 
				 LocalGeneralizedArgs),
	get_list_of_generalizable_terms2(VarIndex, 
	                                 ListOfArgLists, 
					 ListOfGeneralizableTerms),
	generalize_termlists(ListOfGeneralizableTerms, 
	                     Generalization),
	append(LocalGeneralizedArgs, [Generalization], GeneralizedArgs).



look_for_equal_top_symbols(_, [], [0, [], Generalization, []]) :-
	getnewvar(Generalization).

look_for_equal_top_symbols(_, [_], [0, [], Generalization, []]) :-
	getnewvar(Generalization).

look_for_equal_top_symbols(VarCounter, 
	                   ConstraintList,
			   BestSimplification) :-
	
	ConstraintList = [sem_and([_, TermList]) | _],
	get_term(VarCounter, TermList, Term),
	find_duplicates(Term, 
			VarCounter, 
			ConstraintList, 
			ConstraintsWithDuplicates, 
			RestOfConstraintList),
	length(ConstraintsWithDuplicates, L1),
	((L1 == 1 ; var(Term)) ->
	    getnewvar(NewVar),
	    LocalBestSimplification1 = [0, [], NewVar, ConstraintList]
	;
	    compute_generalization(VarCounter, 
	                           ConstraintsWithDuplicates, 
				   Generalization),
	    LocalBestSimplification1 = [VarCounter, 
	                                ConstraintsWithDuplicates,
					Generalization, 
					RestOfConstraintList]
	),

	look_for_equal_top_symbols(VarCounter, 
	                           RestOfConstraintList, 
				   LocalBestSimplification2),

	(is_better_simplification_than(LocalBestSimplification2,
	                               LocalBestSimplification1) ->
	        LocalBestSimplification2 = [VarCounter2, 
		                            ConstraintsWithDuplicates2,
					    Generalization2, 
					    RestOfConstraintList2],
		append(ConstraintsWithDuplicates, 
		       RestOfConstraintList2, 
		       TotalRestOfConstraintList2),
		BestSimplification = [VarCounter2, 
		                      ConstraintsWithDuplicates2,
				      Generalization2, 
				      TotalRestOfConstraintList2]
	    ;   
		BestSimplification = LocalBestSimplification1
	).



% Get the Position-th element of TermList. 
get_term(0, List, _) :-
	!,
	write('ERROR: trying to get 0-th element of the list '),
	writeln(List),
	flush(output),
	fail.
get_term(1, [Term|_], Term) :-
	!.
get_term(Position, [_|TailTermList], Term) :-
	NewPosition is Position - 1,
	get_term(NewPosition, TailTermList, Term).



% Remove the Position-th element of TermList. 
remove_term(0, _, _) :-
	!,
	writeln('ERROR: trying to remove 0-th element of list'),
	flush(output),
	fail.
remove_term(1, [_ | TailTermList], TailTermList) :-
	!.
remove_term(Position, [Term | TailTermList], [Term | NewTailTermList]) :-
	NewPosition is Position - 1,
	remove_term(NewPosition, TailTermList, NewTailTermList).
	


find_duplicates(_, _, [], [], []).

find_duplicates(Term, 
	        Position, 
		ConstraintList, 
		ConstraintsWithDuplicates, 
		RestOfConstraintList) :-

	ConstraintList = [HeadConstraint | TailConstraints],
	HeadConstraint = sem_and([_, TermList]),
	get_term(Position, TermList, T),

	find_duplicates(Term,
			Position,
			TailConstraints,
			LocalConstraintsWithDuplicates,
			LocalRestOfConstraintList),

	(term_comparison(Term, T) ->
	    ConstraintsWithDuplicates = [HeadConstraint | LocalConstraintsWithDuplicates],
	    RestOfConstraintList      = LocalRestOfConstraintList
	;
	    ConstraintsWithDuplicates = LocalConstraintsWithDuplicates,
	    RestOfConstraintList      = [HeadConstraint | LocalRestOfConstraintList]
	).



term_comparison(Term1, Term2) :-
	((cmd_arg(equconstr, (_, _, decomposeVLC, _))
         ;
	  cmd_arg(equconstr, (_, _, decomposeLVC, _))
         ;
	  cmd_arg(equconstr, (_, _, decomposeLCV, _))
         ) ->
	    equal_top_symbols(Term1, Term2)
	;
	    equal_terms(Term1, Term2)
	).



% Succeeds if both terms are variables,
% because these can be renamed such that they are equal.
equal_top_symbols(Term1, Term2) :-
	var(Term1),
	var(Term2).

% Succeeds for equal constants or terms with equal functors.
equal_top_symbols(Term1, Term2) :-
	\+ var(Term1),
	\+ var(Term2),
	functor(Term1, Functor, Arity),
	functor(Term2, Functor, Arity).



% equal ground terms
equal_terms(Term1, Term2) :-
	\+ nonground(Term1),       % Term1 is ground
	\+ nonground(Term2),       % Term2 is ground
	!,
	Term1 == Term2.

% variables are considered equal since they can be renamed
equal_terms(Term1, Term2) :-
	var(Term1),
	var(Term2).

% equal non-variable terms 
% (for efficiency: terms have to be non-ground)
equal_terms(Term1, Term2) :-
	\+ var(Term1),
	\+ var(Term2),
	nonground(Term1),
	nonground(Term2),
	!,
	functor(Term1, Functor, Arity),
	functor(Term2, Functor, Arity),
	Term1 =.. [Functor | Args1],
	Term2 =.. [Functor | Args2],
	compare_termlists(Args1, Args2).

compare_termlists([], []).

compare_termlists(List1, List2) :-
	List1 = [Head1 | Tail1],
	List2 = [Head2 | Tail2],
	equal_terms(Head1, Head2),
	compare_termlists(Tail1, Tail2).



apply_simplification(0, _, _, _, _, _) :-
	!,
	writeln('ERROR: apply_simplification called for not simplifyable constraints'),
	flush(output),
	fail.

apply_simplification(_, _, _, none, _, _) :-
	!,
	writeln('ERROR: apply_simplification called for not simplifyable constraints'),
	flush(output),
	fail.

apply_simplification(Position, 
	             SimplifiableConstraints, 
		     Generalization,
		     equaltopsymbols,
		     ElemConstr,
		     SimplifiedConstraints) :-

	SimplifiableConstraints = [sem_and([VarList, _]) | _],
	get_term(Position, VarList, Var),
	ElemConstr = [[Var], [Generalization]],

	term_variables(Generalization, GeneralizationVarList),
	replace_elem_constr_by_generalizer(SimplifiableConstraints, 
	                                   Position,
					   Generalization,
					   GeneralizationVarList,
					   SimplifiedConstraints),
	rename_varlist(GeneralizationVarList, RenamedVarList),
	term_variables(Generalization, RenamedVarList).


apply_simplification(Position, 
	             SimplifiableConstraints, 
		     RepeatedPosition,
		     repeatedrows,
		     ElemConstr,
		     SimplifiedConstraints) :-

	SimplifiableConstraints = [sem_and([VarList, _]) | _],
	get_term(Position, VarList, Var),
	get_term(RepeatedPosition, VarList, VarAtRepeatedPosition),
	ElemConstr = [[Var], [VarAtRepeatedPosition]],

	% remove Position instead of RepeatedPosition, 
	% because RepeatedPosition < Position
	remove_repeated_row(SimplifiableConstraints, 
	                    Position,
			    SimplifiedConstraints).



remove_repeated_row([], _, []).

remove_repeated_row(SimplifiableConstraints, 
	            RepeatedPosition,
		    SimplifiedConstraints) :-
	SimplifiableConstraints = [sem_and([VarList, TermList]) | Tail],
	remove_term(RepeatedPosition, VarList, NewVarList),
	remove_term(RepeatedPosition, TermList, NewTermList),
	remove_repeated_row(Tail, 
	                    RepeatedPosition,
			    LocalSimplifiedConstraints),
	SimplifiedConstraints = [sem_and([NewVarList, NewTermList]) | LocalSimplifiedConstraints].
	



look_for_repeated_rows(VarCounter, 
	               ConstraintList, 
		       Simplification) :-

	compute_lists_of_repeated_positions(VarCounter,
	                                    ConstraintList,
					    ListsOfRepeatedPositions),

	compute_union(ListsOfRepeatedPositions, 
	              UnionedListsOfRepeatedPositions),
	compute_best_repeated_position(ListsOfRepeatedPositions,
	                               UnionedListsOfRepeatedPositions,
				       NrOfRepetitions,
	                               RepeatedPosition),

	(NrOfRepetitions =< 1 ->
	    Simplification = [0, [], 0, ConstraintList]
	;
	    compute_repeated_rows(VarCounter, 
	                          RepeatedPosition,
				  ConstraintList,
				  ConstraintsWithRepeatedRows,
				  RestOfConstraintList),
	    Simplification = [VarCounter, 
	                      ConstraintsWithRepeatedRows,
			      RepeatedPosition, 
			      RestOfConstraintList]).



compute_lists_of_repeated_positions(_, [], []).

compute_lists_of_repeated_positions(VarCounter,
	                            ConstraintList,
				    ListsOfRepeatedPositions) :-
	ConstraintList = [sem_and([_, TermList]) | TailOfConstraintList],
	get_term(VarCounter, TermList, Term),
	StartPosition is VarCounter - 1,
	compute_repetitions_of_term(Term, 
                                    TermList, 
				    StartPosition, 
				    ListOfRepetitions),
	compute_lists_of_repeated_positions(VarCounter,
	                                    TailOfConstraintList,
					    LocalListsOfRepeatedPositions),
	ListsOfRepeatedPositions = [ListOfRepetitions | LocalListsOfRepeatedPositions].



compute_repetitions_of_term(Term, 
                            TermList, 
			    CheckPosition, 
			    ListOfRepetitions) :-
	CheckPosition > 0,
	get_term(CheckPosition, TermList, CheckTerm),
	(Term == CheckTerm ->
	    ListOfRepetitions = [CheckPosition | LocalListOfRepetitions]
	;
	    ListOfRepetitions = LocalListOfRepetitions
	),
	NewCheckPosition is CheckPosition - 1,
	compute_repetitions_of_term(Term, 
                                    TermList, 
				    NewCheckPosition, 
				    LocalListOfRepetitions).

compute_repetitions_of_term(_, _, CheckPosition, []) :-
	CheckPosition =< 0.
	


compute_union([], []).
compute_union([List], List).
compute_union(ListOfLists, List) :-
	ListOfLists = [Head | Tail],
	compute_union(Tail , LocalList),
	union(Head, LocalList, List).



compute_best_repeated_position([], _, _, _) :-
	!,
	writeln("ERROR: ListsOfRepeatedPositions is empty"),
	fail.

compute_best_repeated_position(_, [], 0, 0).

compute_best_repeated_position(ListsOfRepeatedPositions,
	                       UnionedListsOfRepeatedPositions,
			       NrOfRepetitions,
	                       RepeatedPosition) :-
	UnionedListsOfRepeatedPositions = [Head | Tail],
	compute_occurrences(Head, 
	                    ListsOfRepeatedPositions, 
			    NrOfOccurrences),
	compute_best_repeated_position(ListsOfRepeatedPositions,
	                               Tail,
				       LocalNrOfRepetitions,
				       LocalRepeatedPosition),
	(LocalNrOfRepetitions > NrOfOccurrences ->
	    NrOfRepetitions  = LocalNrOfRepetitions,
	    RepeatedPosition = LocalRepeatedPosition
	;
	    NrOfRepetitions  = NrOfOccurrences,
	    RepeatedPosition = Head
	).



compute_occurrences(_, [], 0).

compute_occurrences(Term, 
	            ListsOfTermLists, 
		    NrOfOccurrences) :-
	ListsOfTermLists = [Head | Tail],
	compute_occurrences(Term, 
	                    Tail, 
			    LocalNrOfOccurrences),
	(occurs(Term, Head) ->
	    NrOfOccurrences is LocalNrOfOccurrences + 1
	;
	    NrOfOccurrences is LocalNrOfOccurrences
	).



compute_repeated_rows(_, _, [], [], []).

compute_repeated_rows(VarCounter, 
	              RepeatedPosition,
		      ConstraintList,
		      ConstraintsWithRepeatedRows,
		      RestOfConstraintList) :-
	ConstraintList = [Head | Tail],
	compute_repeated_rows(VarCounter, 
	                      RepeatedPosition,
			      Tail,
			      LocalConstraintsWithRepeatedRows,
			      LocalRestOfConstraintList),
	Head = sem_and([_, TermList]),
	get_term(VarCounter, TermList, Term),
	get_term(RepeatedPosition, TermList, RepeatedTerm),
	(Term == RepeatedTerm ->
	    ConstraintsWithRepeatedRows = [Head | LocalConstraintsWithRepeatedRows],
	    RestOfConstraintList        = LocalRestOfConstraintList
	;
	    ConstraintsWithRepeatedRows = LocalConstraintsWithRepeatedRows,
	    RestOfConstraintList        = [Head | LocalRestOfConstraintList]
	).





get_cascaded_disjunction([], _) :-
	!,
	writeln('ERROR: get_cascaded_disjunction called for empty list of constraints'),
	flush(output),
	fail.

get_cascaded_disjunction([sem_and(List)], OutList) :-
	List = [VarList, TermList],
	make_elem_constrs_triv_sat(VarList, TermList, OutVarList, OutTermList),
	OutList = [OutVarList, OutTermList].

get_cascaded_disjunction([sem_and(List1), sem_and(List2)], OutConstr) :-
	List1 = [VarList1, TermList1],
	List2 = [VarList2, TermList2],
	make_elem_constrs_triv_sat(VarList1, TermList1, OutVarList1, OutTermList1),
	make_elem_constrs_triv_sat(VarList2, TermList2, OutVarList2, OutTermList2),
	OutList1 = [OutVarList1, OutTermList1],
	OutList2 = [OutVarList2, OutTermList2],
	length(VarList1, L1),
	length(VarList2, L2),
	(L2 > L1 ->
	    OutConstr = sem_or(OutList2, OutList1)
	;
	    OutConstr = sem_or(OutList1, OutList2)
	).

get_cascaded_disjunction([sem_and(List) | TailConstraints], 
	                 sem_or(NewTailConstraint, OutList)) :-
	List = [VarList, TermList],
	make_elem_constrs_triv_sat(VarList, TermList, OutVarList, OutTermList),
	OutList = [OutVarList, OutTermList],
	get_cascaded_disjunction(TailConstraints, NewTailConstraint).



replace_elem_constr_by_generalizer([], _, _, _, []).

replace_elem_constr_by_generalizer([], _, Generalization, _, []) :-
	var(Generalization),
	!,
	writeln('ERROR: Generalization is a variable'),
	flush(stdout),
	fail.

replace_elem_constr_by_generalizer(SimplifiableConstraints, 
	                           Position,
				   Generalization,
				   GeneralizationVarList,
				   SimplifiedConstraints) :-
	SimplifiableConstraints = [sem_and([VarList, TermList]) | TailInConstraints],
	replace_term_by_termlist(Position, VarList, GeneralizationVarList, NewVarList),

	copy_term(Generalization, LocalGeneralization),
	term_variables(LocalGeneralization, LocalGeneralizationVarList),
	get_term(Position, TermList, Term),
	never_failing_unify(LocalGeneralization, Term),
	replace_term_by_termlist(Position, TermList, LocalGeneralizationVarList, NewTermList),

	replace_elem_constr_by_generalizer(TailInConstraints, 
	                                   Position,
					   Generalization,
					   GeneralizationVarList,
					   TailOutConstraints),
	SimplifiedConstraints = [sem_and([NewVarList, NewTermList]) | TailOutConstraints].



never_failing_unify(Term1, Term2) :-
	not_unify(Term1, Term2),
	!,
	write('ERROR: '),
	write(Term1), write(' and '), write(Term2), writeln(' are not unifiable'),
	flush(output),
	fail.

never_failing_unify(Term1, Term2) :-
	Term1 = Term2.


replace_term_by_termlist(0, _, _, _) :-
	!,
	writeln('ERROR: trying to replace 0-th element of list'),
	flush(output),
	fail.
	
replace_term_by_termlist(1, TermList, NewHeadList, NewTermList) :-
	!,
	TermList = [_ | Tail],
	append(NewHeadList, Tail, NewTermList).

replace_term_by_termlist(Position, TermList, NewHeadList, NewTermList) :-
	TermList = [Head | Tail],
	NewPosition is Position - 1,
	replace_term_by_termlist(NewPosition, Tail, NewHeadList, LocalTermList),
	NewTermList = [Head | LocalTermList].




compose(Junctor, Constraint1, Constraint2, Depth1, Depth2, OutConstraint) :-

	(Junctor = sem_and ->
	    
	    % At AND junctors place constraints with lower depth 
	    % before constraints with higher depth
	    % (the easiest first)
	    (Depth1 > Depth2 ->
		OutConstraint =.. [Junctor, Constraint2, Constraint1];
		OutConstraint =.. [Junctor, Constraint1, Constraint2]);
	    
	    % At OR junctors place constraints with higher depth 
	    % before constraints with lower depth
	    % (the most complicated first)
	    (Depth1 < Depth2 ->
		OutConstraint =.. [Junctor, Constraint2, Constraint1];
		OutConstraint =.. [Junctor, Constraint1, Constraint2])).




%************* PROCEDURES WHICH ARE CURRENTLY NOT USED ********************

get_dictionary(_, [], []).

get_dictionary(Term, InDictionary, OutDictionary) :-
	InDictionary = [Head | Tail],
	get_dictionary(Term, Tail, LocalOutDictionary),

	term_variables(Head, VarList),
	length(VarList, L),
	(L == 0 ->
	    OutDictionary = LocalOutDictionary
	;
	    (L == 1 ->
		VarList = [Var],
		(occurs(Var, Term) ->
		    OutDictionary = [Head | LocalOutDictionary]
		;
		    OutDictionary = LocalOutDictionary
		)
	    ;
	        % L > 1
	        !,
	        writeln('ERROR in dictionary'),
		flush(output),
		fail
	    )
	).



lookup_in_dictionary(Right, [], _) :-
	!,
	write('ERROR in lookup_in_dictionary: variable '),
	write(Right),
	writeln(' not found.'),
	flush(output),
	fail.
	

lookup_in_dictionary(Right, [L = R | _], Left) :-
	Right == R,
	!,
	Left = L.

lookup_in_dictionary(Right, [_ | TailOfDictionary], Left) :-
	lookup_in_dictionary(Right, TailOfDictionary, Left).
