% -*- prolog -*-

% ---------------------- Simplification -------------------
% ---------------------------------------------------------


% -----------------------------------------------------------------
% simplify_constraint(InConstraint, OutConstraint)
% Try to simplify this Constraint by reducing duplicates, e.g.:
% ( A = a & B = b1 & C = c1) v ( A = a & B = b2 & C = c2) v (A = a & B = b1 & C = c3) v (A = aaa & B = b1 & C = c1) 
% == or(and([[A, B, C],[a, b1, c1]]), and([[A, B, C],[a, b2, c2]]), and([[A, B, C],[a, b1, c3]]), and([[A, B, C],[aaa, b1, c1]]))
% can be simplified to 
% ((A = a) & ((B = b1 & C = c1) v (B = b2 & C = c2) v (B = b1 & C = c3))) v (A = aaa & B = b1 & C = c1)
% which can be simplified furthermore to
% ((A = a) & (((B = b1) & (C = c1 v C = c3)) v (B = b2 & C = c2))) v (A = aaa & B = b1 & C = c1)
% Note that it's not possible to find duplicates among conjunctions!
% -----------------------------------------------------------------

simplify(InConstraint, OutConstraint) :-
	simplify_sub(InConstraint, OutConstraintSub, Modified),
	% Has InConstraint been modified? If yes, try to simplify once more
	(Modified == modified -> simplify(OutConstraintSub, OutConstraint);
				 % If the constraint contains junctors with arity 1 now,
	                         % remove these junctors
	                         cleanup_junctors(OutConstraintSub, OutConstraint)).


cleanup_junctors(Constr, OutConstr) :-
	Constr =.. [Junctor|Arguments],
	% Junctor != '.'
	cleanup_junctors_list(Arguments, CleanedArguments),
	length(CleanedArguments, L),
	((L \== 1; constraint_with_arity_one(CleanedArguments)) -> 
	    OutConstr =.. [Junctor|CleanedArguments];
	    OutConstr =.. CleanedArguments).

cleanup_junctors_list([],[]).
cleanup_junctors_list([F1|R1], [F2|R2]) :-
	cleanup_junctors_sub(F1, F2),
	cleanup_junctors_list(R1, R2).

cleanup_junctors_sub(Constr, OutConstr) :-
	Constr =.. [Junctor|Arguments],
	(Junctor == '.' -> OutConstr = Constr;
	    cleanup_junctors_list(Arguments, CleanedArguments),
	    praefix_junctor(CleanedArguments, Junctor, OutConstr)).

praefix_junctor([[LS, RS]], _, [LS, RS]) :-
	length(LS, 1), !.
praefix_junctor([[LS, RS]], Junctor, Out) :- !,
	Out =.. [Junctor|[LS, RS]].
praefix_junctor(Args, _, Out) :-
	length(Args, 1), !,
	Out =.. Args.
praefix_junctor(Args, Junctor, Out) :-
	Out =.. [Junctor|Args].

constraint_with_arity_one([[LS, _]]) :-
	length(LS, 1).


simplify_sub(InConstraint, OutConstraint, Modified) :-
	InConstraint =.. [Junctor | SubConstraints],
	(
	    % If the Junctor is an 'and', we can simplify only it's arguments
	    Junctor == and,
	       simplify_subconstraints(SubConstraints, SimplifiedSubConstraints, Modified),
	       OutConstraint =.. [and|SimplifiedSubConstraints];
            % If the Junctor is an 'or', group the Subconstraints by duplicates and
            % delete them
	    Junctor == or,
	       group_subconstraints(SubConstraints, SimplifiableSubConstraints),
	       delete_duplicates(SimplifiableSubConstraints, SimplifiedSubConstraints, Modified),
	       OutConstraint =.. [or|SimplifiedSubConstraints];
	    % An elementary constraint can't be simplified
	    Junctor == '.',
	       OutConstraint = InConstraint
	).

	       
simplify_subconstraints([], [], _).
simplify_subconstraints([SubConstr| TailSubConstrs], [SimplifiedSubConstr | TailSimplifiedSubConstrs], Modified) :-
	simplify_sub(SubConstr, SimplifiedSubConstr, Modified),
	simplify_subconstraints(TailSubConstrs, TailSimplifiedSubConstrs, Modified).
	       

% ----------------------------------------------------------
% group_subconstraints(SubConstraints, GroupedSubConstraints)
%
% Take the first constraint from SubConstraints and find all constraints in SubConstraints
% having an duplicate in common him. Memorize these constraints and the duplicate in GroupedSubConstraints.
% Proceed recursively.
% ----------------------------------------------------------

group_subconstraints([], []).
group_subconstraints([SubConstraint|SubConstraints], [HeadSimplifiable|TailSimplifiable]) :-
	collect_duplicates(SubConstraint, SubConstraints, RestSubConstraints, [], HeadSimplifiable, _),
	group_subconstraints(RestSubConstraints, TailSimplifiable).

% if SC and FirstSC have an duplicate in common, memorize FirstSC
collect_duplicates(SC, [FirstSC|RestSCs], NotSimplifiableSCs, FoundSimplifiableSCs, SimplifiableSCs, Duplicate) :-
	contains_duplicate(SC, FirstSC, Duplicate),!,
	collect_duplicates(SC, RestSCs, NotSimplifiableSCs, [FirstSC|FoundSimplifiableSCs], SimplifiableSCs, Duplicate).
% if not, save FirstSC for later purposes and proceed
collect_duplicates(SC, [FirstSC|RestSCs], [FirstSC|NotSimplifiableSCs], FoundSimplifiableSCs, SimplifiableSCs, Duplicate) :-
	collect_duplicates(SC, RestSCs, NotSimplifiableSCs, FoundSimplifiableSCs, SimplifiableSCs, Duplicate).
% we are ready. Also memorize SC himself and the duplicate
collect_duplicates(SC, [], [], FoundSimplifiableSCs, (Duplicate, [SC|FoundSimplifiableSCs]), Duplicate).


contains_duplicate(and([Vars1, RightSides1]), and([Vars2, RightSides2]), D) :-
	% Do we already know which duplicate to search for?
	var(D) -> contains_duplicate_sub(Vars1, RightSides1, Vars2, RightSides2, D);
	          contains_duplicate_det(Vars2, RightSides2, D).

contains_duplicate_sub(_, _, [], [], []) :- !, fail.
contains_duplicate_sub([], [], _, _, _)  :- !, fail.
contains_duplicate_sub([V1|_], [RightSide1|_],[V2|_], [RightSide2|_], [[V1], [RightSide1]]) :-
	V1 == V2,
	RightSide1 == RightSide2, !.
contains_duplicate_sub([V1|RestV1], [RS1|RSs1], [V2|RestV2], [RS2|RSs2], D) :-
	% search in both sides!
	contains_duplicate_sub([V1|RestV1], [RS1|RSs1], RestV2,      RSs2,       D);
	contains_duplicate_sub(RestV1,       RSs1,      [V2|RestV2], [RS2|RSs2], D).

contains_duplicate_det([], [], _) :- !, fail.
contains_duplicate_det([CurV|_], [CurRS|_], [[V],[RS]]) :-
	CurV == V,
	CurRS == RS, !.
contains_duplicate_det([_|Vs], [_|RSs], D) :-
	contains_duplicate_det(Vs, RSs, D).


% ---------------------------------------------------------------------------------
% delete_duplicates(SimplifiableSubConstraints, SimplifiedSubConstraints, Modified)
% 
% delete the duplicates we found and also recursively simplify the new components
% Set Modified = modified, if a deletion was possible
% ---------------------------------------------------------------------------------

delete_duplicates([],[],_).
delete_duplicates([([], [SC])|RestSimplifiableSCs], [SubSC|SimplifiedSCs], Modified) :- 
	simplify_sub(SC, SubSC, Modified),
	delete_duplicates(RestSimplifiableSCs, SimplifiedSCs, Modified).
delete_duplicates([([[V], [RS]], SubConstraints)|RestSimplifiableSCs], [SubSC|SimplifiedSCs], modified) :-
	delete_duplicates_sub(V, SubConstraints, S1SubConstraints),
	simplify_subconstraints(S1SubConstraints, S2SubConstraints, modified),
	length(S2SubConstraints, L),
	compose_subconstraints(S2SubConstraints, L, [[V],[RS]], SubSC),
	delete_duplicates(RestSimplifiableSCs, SimplifiedSCs, modified).

compose_subconstraints([],  0, SC,  SC) :- !.
compose_subconstraints(SC1, 1, SC2, and(SC2, SC)) :-
	!, SC =.. SC1.
compose_subconstraints(SC1, _, SC2, and(SC2, SC)) :-
	SC =.. [or|SC1].

delete_duplicates_sub(_,[],[]).
delete_duplicates_sub(V, [and([[V2], _])|R1], R2) :-
	V == V2, !,
	delete_duplicates_sub(V, R1, R2).
delete_duplicates_sub(V, [and([LS, RS])|Rest1], [and([LSOut, RSOut])|Rest2]) :-
	delete_duplicates_SC(V, LS, RS, [], [], LSOut, RSOut),
	delete_duplicates_sub(V, Rest1, Rest2).

delete_duplicates_SC(V, [V1|RestV1], [_|RestRS1], AccV, AccRS, Vs2, RSs2) :-
	V == V1, !,
	append(AccV, RestV1, Vs2),
	append(AccRS, RestRS1, RSs2).
delete_duplicates_SC(V, [V1|RestV1], [RS1|RestRS1], AccV, AccRS, Vs2, RSs2) :-
	delete_duplicates_SC(V, RestV1, RestRS1, [V1|AccV], [RS1|AccRS], Vs2, RSs2).











